import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class JADE(Optimizer):
    """Adaptive Differential Evolution (JADE).

    Reference
    ---------
    Zhang, J., and Sanderson, A. C. 2009.
    JADE: Adaptive differential evolution with optional external archive.
    IEEE Transactions on Evolutionary Computation, 13(5), 945–958.
    https://doi.org/10.1109/TEVC.2009.2014613
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self.mu_cr = 0.5
        self.mu_f = 0.5
        self.archive = np.empty((0, self.ndim_problem))
        self.p = 0.2
        self.c = 0.2  # 0.05 - 0.2
        self._n_generations = 0

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(
            self.initial_lower_boundary, self.initial_upper_boundary,
            (self.n_individuals, self.ndim_problem))  # initial point
        yy = []
        for i in range(self.n_individuals):
            yy.append(self._evaluate_fitness(x[i], args))
        return x, yy

    def iterate(self, args=None, x=None, fitness=None):
        cr = self.rng_optimization.normal(self.mu_cr, 0.1, (self.n_individuals,))
        f = self.rng_optimization.normal(self.mu_f, 0.1, (self.n_individuals,))

        # The union of the current population and the archive
        x_union = np.copy(x)
        if len(self.archive) > 0:
            x_union = np.vstack((x_union, self.archive))

        # r1 and r2
        n_ind = np.arange(self.n_individuals)
        r1 = np.array([self.rng_optimization.choice(np.setdiff1d(n_ind, np.array([i]))) for i in n_ind])
        r2 = np.array([self.rng_optimization.choice(
            np.setdiff1d(np.arange(len(x_union)), np.union1d(np.array([i]), r1[i]))) for i in n_ind])

        # top 100p%
        yy = fitness[-self.n_individuals:]
        sort_index = np.argsort(yy)[:int(self.p * self.n_individuals)]
        x_best = x[self.rng_optimization.choice(sort_index, (self.n_individuals,))]

        # Mutation
        f_m = np.tile(f.reshape(self.n_individuals, 1), (1, self.ndim_problem))
        v = x + f_m * (x_best - x) + f_m * (x[r1] - x_union[r2])

        # Crossover
        x_crossover = np.zeros((self.n_individuals, self.ndim_problem))
        r = self.rng_optimization.random((self.n_individuals, self.ndim_problem))
        for i in range(self.n_individuals):
            for j in range(self.ndim_problem):
                if (j == self.rng_optimization.integers(0, self.ndim_problem)) or (r[i][j] < cr[i]):
                    x_crossover[i, j] = v[i, j]
                else:
                    x_crossover[i, j] = x[i, j]

        # Selection
        s_f = []  # the set of all successful crossover probabilities
        s_cr = []  # the set of all successful mutation factors
        x_new = np.zeros((self.n_individuals, self.ndim_problem))
        for i in range(self.n_individuals):
            y_i = self._evaluate_fitness(x_crossover[i])
            if yy[i] <= y_i:
                x_new[i] = x[i]
                fitness.append(yy[i])
            else:
                x_new[i] = x_crossover[i]
                fitness.append(y_i)
                self.archive = np.vstack((self.archive, x_crossover[i]))
                s_cr.append(cr[i])
                s_f.append(f[i])

        # update
        self.mu_cr = (1 - self.c) * self.mu_cr + self.c * np.mean(np.array(s_cr))
        if np.sum(np.array(s_f)) != 0:
            self.mu_f = (1 - self.c) * self.mu_f + self.c * np.sum(np.power(np.array(s_f), 2)) / np.sum(np.array(s_f))
        else:
            self.mu_f = (1 - self.c) * self.mu_f
        return x_new, fitness

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        x, y = self.initialize(args)
        fitness.extend(y)
        while True:
            x, fitness = self.iterate(args, x, fitness)

            # Update archive
            if len(self.archive) > self.n_individuals:
                self.archive = self.archive[
                    self.rng_optimization.choice(np.arange(len(self.archive)), (self.n_individuals,))]

            if self._check_terminations():
                break
            self._n_generations += 1
        results = self._collect_results(fitness)
        return results



import numpy as np

from pypop7.optimizers.cem.cem import CEM
import torch
from lml import LML


class DCEM(CEM):
    """Differentiable Cross-Entropy Method
    Reference
    ---------------
    B. Amos, D. Yarats
    The Differentiable Cross-Entropy Method
    Proceedings of the 37th International Conference on Machine Learning, PMLR 119:291-302, 2020.
    http://proceedings.mlr.press/v119/amos20a.html
    """
    def __init__(self, problem, options):
        CEM.__init__(self, problem, options)
        self.lml_verbose = options.get('lml_verbose')
        self.lml_eps = options.get('lml_eps')

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        mean = self._initialize_mean(is_restart)
        return mean, x, y

    def iterate(self, mean, x, y):
        for i in range(self.n_individuals):
            x[i] = self.rng_optimization.normal(mean, self.sigma, self.ndim_problem)
            y[i] = self._evaluate_fitness(x[i])
            if self._check_terminations():
                return x, y
        return x, y

    def update_distribution(self, x, y):
        mean_y = np.mean(y)
        std_y = np.std(y)
        _y = (y - mean_y) / (std_y + 1e-10)
        _y = _y.reshape(self.n_individuals, 1)
        _y = torch.from_numpy(_y)
        I = LML(N=self.n_parents, verbose=self.lml_verbose, eps=self.lml_eps)(_y)
        I = I.numpy()
        x_I = I * x
        mean = np.mean(x_I, axis=0)
        self.sigma = np.sqrt(np.mean(I * (x - mean)**2, axis=0))
        return mean

    def optimize(self, fitness_function=None):
        fitness = CEM.optimize(self, fitness_function)
        mean, x, y = self.initialize()
        while True:
            x, y = self.iterate(mean, x, y)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            mean = self.update_distribution(x, y)
        results = self._collect_results(fitness, mean)
        return results



"""Repeat the following paper for `FEP`:
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.ep.fep import FEP


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12 * np.ones((ndim_problem,)),
               'upper_boundary': 5.12 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5000 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 22.88404827530138
    # vs 4.6e-2 (from the original paper)

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32 * np.ones((ndim_problem,)),
               'upper_boundary': 32 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 1.9075267778220923
    # vs 1.8e-2 (from the original paper)

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.20284227582986794
    # vs 5.7e-4 (from the original paper)

    problem = {'fitness_function': step,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 47.0
    # vs 0 (from the original paper)

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -30*np.ones((ndim_problem,)),
               'upper_boundary': 30*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 20000 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 19.653232176391665
    # vs mean 5.06 std 5.87 (from the original paper)



import numpy as np

from pypop7.optimizers.ds.ds import DS


class POWELL(DS):
    """Powell's Method(POWELL).
    Reference
    ---------
    Kochenderfer, M.J. and Wheeler, T.A., 2019.
    Algorithms for optimization.
    MIT Press.
    https://algorithmsbook.com/optimization/files/chapter-7.pdf
    (See Algorithm 7.3 (Page 102) for details.)
    M. J. D. Powell
    An Efficient Method for Finding the Minimum of a Function
    of Several Variables Without Calculating Derivative
    Comput.J. 7(2)155-162(1964)
    https://academic.oup.com/comjnl/article-abstract/7/2/155/335330?redirectedFrom=fulltext&login=false
    """
    def __init__(self, problem, options):
        DS.__init__(self, problem, options)
        self.q = 0.1

    def initialize(self, args=None, is_restart=False):
        x = self._initialize_x(is_restart)  # initial point
        y = self._evaluate_fitness(x, args)  # fitness
        u = np.zeros((self.ndim_problem, self.ndim_problem))
        for i in range(self.ndim_problem):
            u[i][i] = 1.0
        return x, y, u

    def line_search(self, x=None, y=None, d=None, args=None):
        xx, yy = np.copy(x), np.copy(y)
        for sgn in [-1, 1]:
            while True:
                if self._check_terminations():
                    return xx, yy
                x_new = xx + sgn * self.q * d
                y_new = self._evaluate_fitness(x_new)
                if y_new < yy:
                    xx = x_new
                    yy = y_new
                else:
                    break
        return xx, yy

    def iterate(self, args=None, x=None, y=None, u=None):
        xx, yy = np.copy(x), np.copy(y)
        for i in range(self.ndim_problem):
            if self._check_terminations():
                return xx, yy, u
            d = u[i]
            xx, yy = self.line_search(xx, yy, d)
            if i != self.ndim_problem - 1:
                u[i] = u[i+1]
        d = xx - x
        d /= np.linalg.norm(d)
        u[-1] = d
        xx, yy = self.line_search(xx, yy, d)
        if y - yy < 0.05 * self.fitness_threshold:
            self.q *= 0.1
        else:
            self.q = 0.4 * np.sqrt(y - yy)
        return xx, yy, u

    def optimize(self, fitness_function=None, args=None):
        fitness = DS.optimize(self, fitness_function)
        x, y, u = self.initialize(args)
        fitness.append(y)
        while True:
            x, y, u = self.iterate(args, x, y, u)
            if self.record_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class CEM(Optimizer):
    """Cross-Entropy Method(CE)
    Reference
    -----------
    P. T. de Boer, D. P. Kroese, S. Mannor, R. Y. Rubinstein, (2003)
    A Tutorial on the Cross-Entropy Method
    http://web.mit.edu/6.454/www/www_fall_2003/gew/CEtutorial.pdf
    T. Hoem-de-Mello, R. Y. Rubinstein,
    Estimation of Rare Event Probabilities using Cross-Entropy
    Proceedings of the 2002 Winter Simulation Conference
    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=1172900
    The official python version:
    https://pypi.org/project/cross-entropy-method/
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:
            self.n_individuals = 100
        if self.n_parents is None:
            self.n_parents = 10
        self._n_generations = 0
        self.sigma = np.ones((self.ndim_problem,)) * options.get('sigma')
        self.mean = options.get('mean')  # mean of Gaussian search distribution
        if self.mean is None:  # 'mean' has priority over 'x'
            self.mean = options.get('x')

    def initialize(self, is_restart=False):
        raise NotImplementedError

    def iterate(self, mean, x, y):
        raise NotImplementedError

    def _initialize_mean(self, is_restart=False):
        if is_restart or (self.mean is None):
            mean = self.rng_initialization.uniform(self.initial_lower_boundary,
                                                   self.initial_upper_boundary)
        else:
            mean = np.copy(self.mean)
        return mean

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            best_so_far_y = -self.best_so_far_y if self._is_maximization else self.best_so_far_y
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness, mean=None):
        results = Optimizer._collect_results(self, fitness)
        results['sigma'] = self.sigma
        results['_n_generations'] = self._n_generations
        return results



import numpy as np

from pypop7.optimizers.es.es import ES


class SAES(ES):
    """Self-Adaptation Evolution Strategy (SAES).

    .. note:: `SAES` adapts only the *global* step-size on-the-fly with *small* populations, often resulting in
       slow (and even premature) convergence for large-scale black-box optimization (LSBBO), especially on
       ill-conditioned optimization problems).

       It is **highly recommended** to first attempt other more advanced ES variants for LSBBO. Here we include
       it mainly for *benchmarking* and *theoretical* purpose.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with five particular settings (`keys`):
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`),
                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`, default:
                  `4 + int(3*np.log(self.ndim_problem))`),
                * 'n_parents'     - number of parents (μ: mu), parental population size (`int`, default:
                  `int(self.n_individuals / 2)`),
                * 'eta_sigma'     - learning rate of global step-size (`float`, default:
                  `1 / np.sqrt(2*self.ndim_problem)`).

    Examples
    --------
    Use the ES optimizer `SAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.saes import SAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> saes = SAES(problem, options)  # initialize the optimizer class
       >>> results = saes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
         * Generation 10: best_so_far_y 1.16832e+00, min(y) 1.52815e+01 & Evaluations 60
         * Generation 20: best_so_far_y 7.96885e-02, min(y) 8.59800e+00 & Evaluations 120
         ...
       SAES: 5000, 0.07968852575335955

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring (λ: lambda), offspring population size.
    n_parents     : `int`
                    number of parents (μ: mu), parental population size.
    mean          : `array_like`
                    initial (starting) point, mean of Gaussian search distribution.
    sigma         : `float`
                    initial global step-size (σ), mutation strength (`float`).
    eta_sigma     : `float`
                    learning rate of global step-size.

    References
    ----------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700). ACM.
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    http://www.scholarpedia.org/article/Evolution_strategies

    https://homepages.fhv.at/hgb/downloads/mu_mu_I_lambda-ES.oct
    (See its official Matlab/Octave version.)
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        if self.eta_sigma is None:
            self.eta_sigma = 1 / np.sqrt(2*self.ndim_problem)

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        sigmas = np.ones((self.n_individuals,))  # step-sizes for all offspring
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return x, mean, sigmas, y

    def iterate(self, x=None, mean=None, sigmas=None, y=None, args=None):
        for k in range(self.n_individuals):  # sample offspring population
            if self._check_terminations():
                return x, sigmas, y
            sigmas[k] = self.sigma*np.exp(self.eta_sigma*self.rng_optimization.standard_normal())
            x[k] = mean + sigmas[k]*self.rng_optimization.standard_normal((self.ndim_problem,))
            y[k] = self._evaluate_fitness(x[k], args)
        return x, sigmas, y

    def _restart_initialize(self):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            self.n_individuals *= 2
            self.n_parents = int(self.n_individuals / 2)
            self._n_generations = 0
            self._fitness_list = [np.Inf]
        return is_restart

    def restart_initialize(self, x=None, mean=None, sigmas=None, y=None):
        if self._restart_initialize():
            x, mean, sigmas, y = self.initialize(True)
        return x, mean, sigmas, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, sigmas, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            x, sigmas, y = self.iterate(x, mean, sigmas, y, args)
            if self.record_fitness:
                fitness.extend(y.tolist())
            if self._check_terminations():
                break
            order = np.argsort(y)[:self.n_parents]
            mean = np.mean(x[order], axis=0)
            self.sigma = np.mean(sigmas[order])
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, mean, sigmas, y = self.restart_initialize(x, mean, sigmas, y)
        return self._collect_results(fitness, mean)



"""Repeat the following two papers for `CEP`:
    Yao, X., Liu, Y. and Lin, G., 1999.
    Evolutionary programming made faster.
    IEEE Transactions on Evolutionary Computation, 3(2), pp.82-102.
    https://ieeexplore.ieee.org/abstract/document/771163

    Bäck, T. and Schwefel, H.P., 1993.
    An overview of evolutionary algorithms for parameter optimization.
    Evolutionary Computation, 1(1), pp.1-23.
    https://direct.mit.edu/evco/article-abstract/1/1/1/1092/An-Overview-of-Evolutionary-Algorithms-for

    Note that we first update individual step-sizes and then update offspring for each generation as the
    second paper, in order to obtain fast local convergence.

    Very close performance can be obtained by our code. Therefore, we argue that the repeatability of
    `CEP` can be well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.ep.cep import CEP


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12*np.ones((ndim_problem,)),
               'upper_boundary': 5.12*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5000 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    cep = CEP(problem, options)
    results = cep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 57.707348276918225
    # vs 89 (from the original paper)

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32*np.ones((ndim_problem,)),
               'upper_boundary': 32*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    cep = CEP(problem, options)
    results = cep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.047112659488285136
    # vs 9.2 (from the original paper)

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    cep = CEP(problem, options)
    results = cep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.00021076229028199638
    # vs 2.2e-4 (from the original paper)

    problem = {'fitness_function': step,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    cep = CEP(problem, options)
    results = cep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 9.0
    # vs mean 577.76 std 1125.76 (from the original paper)

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -30*np.ones((ndim_problem,)),
               'upper_boundary': 30*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 20000 * 100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    cep = CEP(problem, options)
    results = cep.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 7.1820167322999175
    # vs mean 6.17 std 13.61 (from the original paper)



import numpy as np

from pypop7.optimizers.eda.eda import EDA
from scipy.linalg import orth


class RPEDA(EDA):
    """Random Projection Estimation of distribution algorithms(RP-EDA)
    Reference
    --------------
    A. Kaban, J. Bootkrajang, R. J. Durrant
    Towards Large Scale Continuous EDA: A Random Matrix Theory Perspective
    GECCO 2013: 383-390
    """
    def __init__(self, problem, options):
        EDA.__init__(self, problem, options)
        self.x = None
        self.k = options.get('k')
        self.typeR = options.get('typeR')
        self.rpmSize = options.get('rpmSize', 1000)
        self.sq_1_d = np.sqrt(1.0 / self.ndim_problem)
        self.sq_d_k = np.sqrt(self.ndim_problem / self.k)
        self.sq_rpm_size = np.sqrt(self.rpmSize)
        assert self.k < self.ndim_problem

    def initialize(self):
        x = np.empty((self.n_individuals, self.ndim_problem))
        x_fit = np.empty((self.n_parents, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            x[i] = self._initialize_x()
            y[i] = self._evaluate_fitness(x[i])
        return x, x_fit, y

    def iterate(self, x_fit, y):
        mean = np.mean(x_fit, axis=0)
        for i in range(self.n_parents):
            x_fit[i] -= mean
        x = np.zeros((self.n_individuals, self.ndim_problem))
        for i in range(self.rpmSize):
            if self.typeR == 'G':
                R = self.rng_optimization.standard_normal((self.ndim_problem, self.k)) * self.sq_1_d
            elif self.typeR == 'g':
                R = self.rng_optimization.standard_normal((self.ndim_problem, self.k)) * self.sq_1_d
            elif self.typeR == 'o':
                R = self.rng_optimization.standard_normal((self.ndim_problem, self.k)) * self.sq_1_d
                R = orth(R)
            elif self.typeR == 'b':
                R = self.rng_optimization.standard_normal((self.ndim_problem, self.k))
                for j in range(self.ndim_problem):
                    for k in range(self.k):
                        if R[j][k] > 0.5:
                            R[j][k] = self.sq_1_d
                        else:
                            R[j][k] = -1 * self.sq_1_d
            elif self.typeR == 's':
                R = self.rng_optimization.standard_normal((self.ndim_problem, self.k))
                for j in range(self.ndim_problem):
                    for k in range(self.k):
                        if R[j][k] > 5.0 / 6:
                            R[j][k] = -2
                        elif R[j][k] > 2.0 / 3:
                            R[j][k] = -1
                        elif R[j][k] > 0:
                            R[j][k] = 0
                        elif R[j][k] == -2:
                            R[j][k] = 1
                R *= np.sqrt(3) * self.sq_1_d
            else:
                raise Exception("Solver not implemented")
            project_matrix = np.dot(x_fit, R)
            cvr = np.zeros((self.k, self.k))
            for j in range(self.n_parents):
                cvr += np.cov(project_matrix[j])
            cvr /= self.n_parents
            xtmp = self.rng_optimization.multivariate_normal(np.zeros((self.k, )), cvr, self.n_individuals)
            x += np.dot(xtmp, np.transpose(R))
        x /= self.rpmSize
        x *= self.sq_d_k * self.sq_rpm_size
        for i in range(self.n_individuals):
            x[i] = np.clip(x[i] + mean, self.lower_boundary, self.upper_boundary)
            if self._check_terminations():
                return x, y
            y[i] = self._evaluate_fitness(x[i])
        return x, y

    def optimize(self, fitness_function=None):
        fitness = EDA.optimize(self, fitness_function)
        x, x_fit, y = self.initialize()
        while True:
            order = np.argsort(y)
            for i in range(self.n_parents):
                x_fit[i] = x[order[i]]
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            x, y = self.iterate(x_fit, y)
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results



import numpy as np
import matplotlib.pyplot as plt
from celluloid import Camera

from pypop7.benchmarks.utils import generate_xyz
from pypop7.optimizers.ds.ds import DS
from pypop7.optimizers.ds.hj import HJ


def plot_contour(func, x, y, levels=None):
    x, y, z = generate_xyz(func, x, y, 500)
    plt.contourf(x, y, z, levels, cmap='bone')
    plt.xlabel('x')
    plt.ylabel('y')


def shi_cd(x):  # fitness function from https://arxiv.org/abs/1610.00040
    return 7 * np.power(x[0], 2) + 6 * x[0] * x[1] + 8 * np.power(x[1], 2)


class HJ1(HJ):
    def iterate(self, args=None, x=None, fitness=None):
        improved, best_so_far_x, best_so_far_y = False, self.best_so_far_x, self.best_so_far_y
        x_data, x_i = np.empty(shape=(4, 2)), 0
        for ii in range(self.ndim_problem):
            for sgn in [-1, 1]:
                if self._check_terminations():
                    return x_data, best_so_far_x
                xx = np.copy(best_so_far_x)
                xx[ii] += sgn * self.sigma
                x_data[x_i] = xx
                x_i += 1
                y = self._evaluate_fitness(xx, args)
                if self.record_fitness:
                    fitness.append(y)
                if y < best_so_far_y:
                    best_so_far_y, improved = y, True
        if not improved:
            self.sigma *= self.gamma  # alpha
        return x_data, best_so_far_x

    def optimize(self, fitness_function=None, args=None):
        fitness = DS.optimize(self, fitness_function)
        x_data, mean_data = [], []
        x, y = self.initialize(args)
        fitness.append(y)
        while True:
            xx, best_so_far_x = self.iterate(args, x, fitness)
            x_data.append(xx)
            mean_data.append(best_so_far_x)
            if self._check_terminations():
                break
            self._n_generations += 1
        results = self._collect_results(fitness)
        results['x_data'] = x_data
        results['mean_data'] = mean_data
        return results


if __name__ == '__main__':
    ndim_problem = 2
    pro = {'fitness_function': shi_cd,
           'ndim_problem': ndim_problem}
    opt = {'fitness_threshold': 1e-10,
           'seed_rng': 0,
           'x': np.array([7., -8.]),
           'sigma': 0.9,
           'record_fitness': True,
           'record_fitness_frequency': 1}
    solver = HJ1(pro, opt)
    res = solver.optimize()
    print(res)
    fig = plt.figure()
    plt.rcParams['font.family'] = 'Times New Roman'
    plt.rcParams['font.size'] = '12'
    camera = Camera(fig)
    for i in range(len(res['mean_data'])):
        plt.xlim([-10, 10])
        plt.ylim([-10, 10])
        plt.xticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plt.yticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plot_contour(shi_cd, [-10, 10], [-10, 10], [0, 10, 100, 500, 1000, 2000])
        plt.scatter(res['x_data'][i][:, 0], res['x_data'][i][:, 1], c='limegreen', s=3)
        plt.scatter(res['mean_data'][i][0], res['mean_data'][i][1], c='magenta', s=12)
        camera.snap()
    animation = camera.animate()
    animation.save('demo_hj.gif')



"""Repeat Fig.5 Function T12 k=3, M=1000 in paper:
    A. Kaban, J. Bootkrajang, R. J. Durrant
    Towards Large Scale Continuous EDA: A Random Matrix Theory Perspective
    GECCO 2013: 383-390
"""
import pickle
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import rosenbrock
from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.eda.rpeda import RPEDA

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function):
    plt.figure()
    result = read_pickle(function, 1000)
    xs = np.arange(len(result))
    plt.yscale('log')
    plt.plot(xs, result, color='r')
    plt.xlim([0, 2e3])
    plt.ylim([1e3, 1e13])
    plt.xticks([200, 400, 600, 800, 1000, 1200, 1400, 1600, 1800, 2000])
    plt.yticks(ticks=[1e4, 1e6, 1e8, 1e10, 1e12])
    plt.xlabel("Generations")
    plt.ylabel("Fitness")
    plt.title(function.capitalize())
    plt.show()


class Fig5(RPEDA):
    def optimize(self, fitness_function=None):
        fitness = Optimizer.optimize(self, fitness_function)
        x, x_fit, y = self.initialize()
        best_fitness = []
        while True:
            order = np.argsort(y)
            for i in range(self.n_parents):
                x_fit[i] = x[order[i]]
            if self.record_fitness:
                fitness.extend(y)
            best_fitness.append(y[order[0]])
            if self._check_terminations():
                break
            x, y = self.iterate(x_fit, y)
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness)
        return results, best_fitness


if __name__ == '__main__':
    ndim_problem = 1000
    for f in [rosenbrock]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -100 * np.ones((ndim_problem,)),
                   'upper_boundary': 100 * np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 6e5,
                   'fitness_threshold': 1e-10,
                   'seed_rng': 0,
                   'k': 3,
                   'rpmSize': 1000,
                   'typeR': 'G',
                   'n_parents': 75,
                   'n_individuals': 300,
                   'verbose_frequency': 200,
                   'record_fitness': False,
                   'record_fitness_frequency': 1}
        solver = Fig5(problem, options)
        results, best_fitness = solver.optimize()
        write_pickle(f.__name__, ndim_problem, best_fitness)
        plot(f.__name__)



import numpy as np

from pypop7.optimizers.rs.rs import RS


class PRS(RS):
    """Pure Random Search (PRS).

    .. note:: `PRS` is one of the *simplest* and *earliest* black-box optimizers. Although recently it
        has been successfully applied in several *relatively low-dimensional* problems (particularly
        `hyper-parameter optimization <https://www.jmlr.org/papers/v13/bergstra12a.html>`_), it generally
        suffers from the famous **curse of dimensionality** for large-scale black-box optimization (LSBBO),
        since its lack of *adaptation*, a highly desirable property for sophisticated search algorithms.

        It is **highly recommended** to first attempt other more advanced methods for LSBBO.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `1000`);
              and with one particular setting (`key`):
                * 'x'     - initial (starting) point (`array_like`).

    Examples
    --------
    Use the Random Search optimizer `PRS` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.rs.prs import PRS
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3 * numpy.ones((2,))}
       >>> prs = PRS(problem, options)  # initialize the optimizer class
       >>> results = prs.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"Pure-Random-Search: {results['n_function_evaluations']}, {results['best_so_far_y']}")
         * Generation 0: best_so_far_y 3.60400e+03, min(y) 3.60400e+03 & Evaluations 1
         * Generation 1000: best_so_far_y 1.14977e-01, min(y) 3.01919e+04 & Evaluations 1001
         * Generation 2000: best_so_far_y 1.14977e-01, min(y) 8.06598e+02 & Evaluations 2001
         * Generation 3000: best_so_far_y 1.14977e-01, min(y) 1.81726e+03 & Evaluations 3001
         * Generation 4000: best_so_far_y 1.14977e-01, min(y) 8.62163e+03 & Evaluations 4001
       Pure-Random-Search: 5000, 0.11497678820610932

    Attributes
    ----------
    x     : `array_like`
            initial (starting) point.

    References
    ----------
    Bergstra, J. and Bengio, Y., 2012.
    Random search for hyper-parameter optimization.
    Journal of Machine Learning Research, 13(2).
    https://www.jmlr.org/papers/v13/bergstra12a.html

    Brooks, S.H., 1958.
    A discussion of random methods for seeking maxima.
    Operations Research, 6(2), pp.244-251.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.6.2.244
    """
    def __init__(self, problem, options):
        RS.__init__(self, problem, options)
        # default: 1 -> uniformly distributed random sampling
        self.sampling_distribution = options.get('sampling_distribution', 1)
        if self.sampling_distribution not in [0, 1]:  # 0 -> normally distributed random sampling
            info = 'For {:s}, only support uniformly or normally distributed random sampling.'
            raise ValueError(info.format(self.__class__.__name__))
        if self.sampling_distribution == 0:
            self.sigma = options.get('sigma')  # initial (global) step-size
            if self.sigma is None:
                raise ValueError('`sigma` should be set.')

    def _sample(self, rng):
        if self.sampling_distribution == 0:
            x = self.x + self.sigma*rng.standard_normal(size=(self.ndim_problem,))
        else:
            x = rng.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        return x

    def initialize(self):
        if self.x is None:
            x = self._sample(self.rng_initialization)
        else:
            x = np.copy(self.x)
        return x

    def iterate(self):  # draw a sample (individual)
        return self._sample(self.rng_optimization)



from abc import ABC

import numpy as np
import math
from scipy.stats import norm

from pypop7.optimizers.es.es import ES


# (2.3.1 Distance Metrix)
def calculate_mahalanobis_distance(yy, x, inv_cov):
    if (yy.ndim == 0) and (inv_cov.ndim == 0):  # yy is a scalar
        return np.sqrt(np.dot(np.dot(yy - x, inv_cov), (yy - x).T))
    elif (yy.ndim == 1) and (inv_cov.ndim == 2):  # yy is a vector
        return np.sqrt(np.dot(np.dot(yy - x, inv_cov), (yy - x).T))
    elif (yy.ndim == 2) and (inv_cov.ndim == 2):  # yy is a two-dimensional matrix
        return np.array([np.sqrt(np.dot(np.dot(y - x, inv_cov), (y - x).T)) for y in yy])
    else:
        raise TypeError(f'{yy} and {inv_cov}.')


class RSCMSA(ES, ABC):
    """Covariance Matrix Self-Adaptation Evolution Strategy with Repelling Subpopulations (RS-CMSA)

    Reference
    ---------
    Ahrari, A., Deb, K., and Preuss, M. 2017.
    Multimodal optimization by covariance matrix self-adaptation evolution strategy with repelling subpopulations.
    Evolutionary Computation, 25(3), 439–471.
    https://doi.org/10.1162/evco_a_00182
    """

    def __init__(self, problem, options):
        ES.__init__(self, problem, options)

        self.tau = np.sqrt(0.5 / self.ndim_problem)  #
        # the default value of the normalized taboo distance, which is assigned to the new members of Archive as well
        # as the subpopulations during the restart.
        self._hat_d_bak = 1  # default value of the normalized taboo distance
        self.tau_hat_d = np.sqrt(1 / self.ndim_problem)  # the learning rate of the normalized taboo distance
        self.alpha_new = 0.5  # target rate of basin identification
        # the size of each taboo region shrinks by (1-self.c_red)
        # A smaller value of self.c_red speeds up reduction of the normalized taboo distances
        self.c_red = np.power(0.99, 1 / self.ndim_problem)
        self.criticality_threshold = 0.01  # criticality threshold for taboo points
        self.n_s = options.get('n_subpopulations', 10)  # number of subpopulations
        self._n_s_bak = np.copy(self.n_s)  # default number of subpopulations
        self.percentile = 25  # Percentile for the normalized taboo distance of the current subpopulations
        self.max_budget = 10  # Max evaluation budget for the hill-valley ( Detect Multimodal) function
        self.distance_metric = options.get('distance_metric', 'Mahalanobis')
        if self.distance_metric not in ['Euclidean', 'Mahalanobis']:
            raise TypeError(f'Distance metric must be "Euclidean" or "Mahalanobis" not {self.distance_metric}.')
        self.best_so_far_x_bak = np.copy(self.best_so_far_x)
        self.best_so_far_y_bak = np.copy(self.best_so_far_y)
        self.n_condition = 1e14

        # the adaptation interval constant for Covariance Matrix
        # (1/self.tau_c) can be interpreted as the learning rate for the Covariance Matrix
        self.tau_c = 1 + self.ndim_problem * (self.ndim_problem + 1) / self.n_parents
        # The n_elt best solutions of this iteration will survive to the next iteration
        self.n_elt = max(1, math.floor(0.15 * self.n_individuals))  # number of elites (2.3.3)
        # average number of iterations used by the subpopulations in the previous restart
        self.average_iteration_subpopulations = 1

    # (2.6.3 Calculation of the Mean Rejection Probability)
    def find_critical_points_descend_index(self, x, sigma, inv_cov, archive_xx, archive_d):
        # Return the index of critical taboo points sorted based on criticality (descend)
        if (archive_xx.ndim == 2) and (archive_xx.size > 0):
            mu = np.sqrt(min(np.abs(np.linalg.eigvals(inv_cov)))) / sigma
            mu = mu * np.sum(np.power(x - archive_xx, 2), 1) - archive_d
            criticality = 1 - 2 * norm.cdf(mu)  # standard normal distribution
            return np.argsort(-1 * criticality)[np.sort(-1 * criticality) < -1 * self.criticality_threshold]
        return np.empty((0,))

    def is_new_basin(self, x1, x4, f1, f4, args=None):
        new_basin, n = False, 0
        if np.sqrt(np.sum(np.power(x1 - x4, 2))) > 0:
            max_f = max(f1, f4)
            direction = x4 - x1  # search direction
            delta = np.sqrt(np.sum(np.power(x4 - x1, 2)))
            direction = direction / delta / 2.618
            x2 = x1 + delta * direction
            f2 = self._evaluate_fitness(x2, args)
            n += 1
            if f2 > max_f:
                new_basin = True
            elif n < self.max_budget:  # Max evaluation budget for the hill-valley ( Detect Multimodal) function
                x3 = x1 + 1.618 * delta * direction
                f3 = self._evaluate_fitness(x3, args)
                n += 1
                if f3 > max_f:
                    new_basin = True
                else:
                    while (n < self.max_budget) and (np.sqrt(np.sum(np.power(x1 - x4, 2))) > 0):
                        if f2 < f3:
                            delta = np.sqrt(np.sum(np.power(x4 - x2, 2))) / 2.618
                            x1, f1 = np.copy(x2), np.copy(f2)
                            x2, f2 = np.copy(x3), np.copy(f3)
                            x3 = x1 + 1.618 * delta * direction
                            f3 = self._evaluate_fitness(x3, args)
                            n += 1
                        elif f2 > f3:
                            delta = np.sqrt(np.sum(np.power(x3 - x1, 2))) / 2.618
                            x4, f4 = np.copy(x3), np.copy(f3)
                            x3, f3 = np.copy(x2), np.copy(f2)
                            x2 = x1 + delta * direction
                            f2 = self._evaluate_fitness(x2, args)
                            n += 1
                        else:
                            delta = np.sqrt(np.sum(np.power(x3 - x2, 2))) / 2.618034
                            x1, f1 = np.copy(x2), np.copy(f2)
                            x4, f4 = np.copy(x3), np.copy(f3)
                            x2 = x1 + delta * direction
                            x3 = x4 - delta * direction
                            f2 = self._evaluate_fitness(x2, args)
                            f3 = self._evaluate_fitness(x3, args)
                            n += 2
                        if max(f2, f3) > max_f:
                            new_basin = True
                            break
        return new_basin

    # (2.4 Restarts with Increasing Population Size)
    def update_population_size(self):
        factor = (self.max_function_evaluations - self.n_function_evaluations) / (
                self.n_individuals * self._n_s_bak * self.average_iteration_subpopulations)
        if factor >= 2:
            self.n_s = np.copy(self._n_s_bak)
            self.n_individuals = 2 * self.n_individuals
        elif factor >= 1:
            self.n_s = np.copy(self._n_s_bak)
            self.n_individuals = math.floor(self.n_individuals * factor)
        else:
            self.n_s = max(1, math.floor(self._n_s_bak * factor))

        self.sigma = np.copy(self._sigma_bak)
        self.n_parents = int(self.n_individuals / 2)
        w_base, w = np.log((self.n_individuals + 1) / 2), np.log(np.arange(self.n_parents) + 1)
        self._w = (w_base - w) / (self.n_parents * w_base - np.sum(w))
        self._mu_eff = 1 / np.sum(np.power(self._w, 2))

        # the adaptation interval constant for Covariance Matrix
        # (1/self.tau_c) can be interpreted as the learning rate for the Covariance Matrix
        self.tau_c = 1 + self.ndim_problem * (self.ndim_problem + 1) / self.n_parents
        # The n_elt best solutions of this iteration will survive to the next iteration
        self.n_elt = max(1, math.floor(0.15 * self.n_individuals))  # number of elites (2.3.3)
        # average number of iterations used by the subpopulations in the previous restart
        self.average_iteration_subpopulations = 1

    # Algorithm 1 (Updating Archive and the normalized taboo distances)
    # (2.5 Adaptation of the Normalized Taboo Distance)
    # Update the archive using the best solutions of the converged subpopulations (yy) and the old archive
    def updating_archive(self, yy, yy_f, archive_xx, archive_f, archive_d, args=None):
        hat_d_0 = self._hat_d_bak  # If archive is empty, then hat_d_0 = 1
        if (archive_f.size > 0) and (archive_d.size > 0):
            if min(yy_f) + self.fitness_threshold < min(archive_f):  # not global minimum, discard them
                archive_xx, archive_f, archive_d = np.empty((0, self.ndim_problem)), np.empty((0,)), np.empty((0,))
            else:
                hat_d_0 = np.percentile(archive_d, self.percentile)

        # Consider only global minima, among the recently generated solutions
        idx = np.arange(yy_f.size)[yy_f <= self.fitness_threshold + np.min(np.hstack((archive_f, yy_f)))]

        # the number of subpopulations that have converged to the corresponding basin
        n_rep = np.empty((len(archive_xx),))
        for i in idx:
            y, f_y = yy[i], yy_f[i]
            # Sort the points in Archive based on their Euclidean distance to y
            archive_xx = archive_xx[np.argsort(np.array([np.linalg.norm(y - x, ord=2) for x in archive_xx]))]
            is_new = 1
            for j in range(len(archive_xx)):
                x, f_x = archive_xx[j], archive_f[j]
                # Check whether share the same basin
                is_basin = True
                if (self.max_function_evaluations - self.n_function_evaluations) > self.max_budget:
                    is_basin = self.is_new_basin(y, x, f_y, f_x, args)
                if is_basin:  # share the same basin
                    if f_y < f_x:  # y is fitter than x
                        archive_xx[j], archive_f[j] = y, f_y
                    n_rep[i] += 1  # (Current restart)
                    is_new = 0  # It is not a new basin
                    break

            if is_new == 1:
                archive_xx = np.vstack((archive_xx, y))
                archive_f = np.hstack((archive_f, f_y))
                n_rep = np.hstack((n_rep, 0))  # (Current restart)
                archive_d = np.hstack((archive_d, hat_d_0))

        # Update the normalized taboo distance based on the (CURRENT) attraction power of the corresponding basin (2.4)
        diff = n_rep - self.alpha_new * (len(yy) / len(archive_xx))
        idx = np.power(-1, diff <= 0)  # A negative value indicates the basin has a weaker attraction power
        archive_d *= np.power(1 + abs(diff), self.tau_hat_d)
        for i in range(len(idx)):
            if idx[i] == -1:
                archive_d[i] = 1 / archive_d[i]  # Reduce the normalized distance
        return archive_xx, archive_f, archive_d

    def generate_subpopulation(self, mean, cov, sigma,  # current subpopulation mean, covariance, and sigma
                               superior_mean, superior_d,  # superior subpopulations; mean and normalized distance
                               archive_xx, archive_f, archive_d,  # all subpopulations; best
                               elt_xx, elt_f, elt_zz, elt_s,  # only current subpopulation; previous elites
                               best_f_ne, median_f_ne, args=None):  # only current subpopulation; history best, median
        """
        elt_n: Number of elite individuals per subpopulation
        elt_xx: initial elite solutions of the subpopulation
        elt_f: initial elite function fitness
        elt_zz: Initial value of the elite variation vector
        elt_s: Initial value of the elite global step sizes
        best_f_ne: History of the best of non-elite solutions for each subpopulation
        med_f_ne: History of the median of non-elite solutions for each subpopulation
        """
        temp_shrinkage_factor = 1  # for temporary shrinkage of the taboo regions
        values, vectors = np.linalg.eig(cov)
        vectors = np.real(vectors)
        values = np.abs(values)  # np.max(v, 0) ??? whether it is right in the author's matlab codes?
        inv_cov = np.matmul(np.matmul(vectors, np.diag(1 / values)), vectors.T)  # Inverse of the matrix
        sqrt_v = np.sqrt(values)

        # Generate taboo points
        taboo_points = np.copy(superior_mean)  # Center of fitter subpopulations
        taboo_points_d = np.copy(superior_d)  # The normalized taboo distance of fitter subpopulations
        for i in range(len(archive_xx)):  # Consider fitter archived points as taboo points
            aa = np.ones((elt_f.size,))[archive_f[i] < elt_f]
            if len(aa) == elt_f.size:  # if elt_f is empty, this will not continue
                taboo_points = np.append(taboo_points, archive_xx[i])
                taboo_points_d = np.append(taboo_points_d, archive_d[i])

        # Determine which taboo points are critical (numpy-ndarray)
        criticality_index = self.find_critical_points_descend_index(mean, sigma, inv_cov, taboo_points, taboo_points_d)

        # Sampling \lamda new solutions
        # the diversity preservation strategy is applied to the sampling stage
        xx = np.zeros((self.n_individuals, self.ndim_problem))
        f = np.zeros((self.n_individuals,))
        sigmas = np.zeros((self.n_individuals,))
        zz = np.copy(xx)
        # Generate \lambda taboo acceptable solutions
        for n in range(self.n_individuals):
            accept = False
            while not accept:
                sigmas[n] = sigma * np.exp(self.rng_optimization.normal(0, 1, 1) * self.tau)
                aa = sqrt_v * self.rng_optimization.normal(0, 1, (self.ndim_problem,))
                zz[n] = np.matmul(vectors, aa)
                xx[n] = mean + sigmas[n] * zz[n]
                if criticality_index.size == 0:
                    accept = True
                for c in criticality_index:
                    d = calculate_mahalanobis_distance(xx[n], taboo_points[c], inv_cov)
                    accept = d > sigma * taboo_points_d[c] * temp_shrinkage_factor
                    if not accept:  # reject
                        temp_shrinkage_factor *= self.c_red  # Temporary shrink the size of the taboo regions
                        break
            f[n] = self._evaluate_fitness(xx[n], args)

        # best for the non-elite solutions
        best_f_ne = np.append(best_f_ne, np.min(f))
        median_f_ne = np.append(median_f_ne, np.median(f))

        # Append the surviving elites from the previous generation (Recombine)
        # At the end of each generation, the solutions (the union of λ recently generated and Nelt surviving solutions
        # from the previous iteration) are sorted in increasing order of the function value.
        # Parameters of the subpopulations are then updated as follows:
        for k in range(len(elt_f)):
            accept = False
            for c in criticality_index:
                d = calculate_mahalanobis_distance(elt_xx[k], taboo_points[c], inv_cov)
                accept = d > sigma * taboo_points_d[c] * temp_shrinkage_factor
                if not accept:
                    break
            if accept:  # The elite was not in a taboo region
                xx = np.vstack((xx, elt_xx[k]))
                f = np.hstack((f, elt_f[k]))
                sigmas = np.hstack((sigmas, elt_s[k]))
                zz = np.vstack((zz, elt_zz[k]))

        # Update parameters of the subpopulation using Equation 6
        idx = np.argsort(f)
        new_mean = np.sum(self._w.reshape(self.n_parents, 1) * xx[idx[:self.n_parents]], 0) / self.n_parents
        n_d = self.ndim_problem
        c = np.sum(np.array([self._w[i] * np.tile(zz[idx[i]].reshape(n_d, 1), (1, n_d)) * np.tile(zz[idx[i]], (n_d, 1))
                             for i in range(self.n_parents)]))
        new_cov = (1 - 1 / self.tau_c) * cov + c / self.tau_c
        new_cov = (new_cov + new_cov.T) / 2  # A symmetric matrix
        numerator = np.exp(np.dot(self._w, np.log(sigmas[:self.n_parents])))
        denominator = np.exp(np.mean(np.log(sigmas)))
        new_sigma = sigma * numerator / denominator

        # Update the elite solutions
        elt_xx = xx[idx[:self.n_elt]]
        elt_f = f[idx[:self.n_elt]]
        elt_s = sigmas[idx[:self.n_elt]]
        elt_zz = zz[idx[:self.n_elt]]
        elt_xx_f_s_zz = (elt_xx, elt_f, elt_s, elt_zz)
        best_xx_f = (xx[idx[0]], f[idx[0]])
        new_mean_cov_sigma = (new_mean, new_cov, new_sigma)
        best_median_f_ne = (best_f_ne, median_f_ne)
        return best_xx_f, new_mean_cov_sigma, elt_xx_f_s_zz, best_median_f_ne

    # Algorithm 2 (2.7 Initialization of Subpopulations)
    def initialize_subpopulations(self, archive_xx, archive_d):
        hat_d = np.copy(self._hat_d_bak)
        if len(archive_d) > 0:
            # Set the default value of the normalized taboo distance based on the distribution of the current values
            # Equal to the $25\text{th}$ percentile of normalized taboo distances of the solutions stored in Archive.
            hat_d = np.percentile(archive_d, self.percentile)
        cov, sigma = np.diag(self.upper_boundary - self.lower_boundary), self._sigma_bak
        inv_cov = np.diag(np.power(self.upper_boundary - self.lower_boundary, -1))
        means = np.empty((self.n_s, self.ndim_problem))

        s, r = 0, 0
        while s < self.n_s:
            if r > 100:
                sigma, r = sigma + self.c_red * sigma, 0  # the size of each taboo region shrinks by (1-self.c_red)
            # Sample uniformly
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
            # for archived points
            accept = True
            for a in range(len(archive_xx)):
                d = calculate_mahalanobis_distance(archive_xx, x, inv_cov)
                if (2 * hat_d + archive_d[a]) * sigma > d:
                    accept = False
                    break
            # for previously sampled points
            if accept:
                for p in range(s):
                    if (3 * hat_d) * sigma > calculate_mahalanobis_distance(means[p], x, inv_cov):
                        accept = False
                        break
            if accept:
                means[s], r = x, 0
                s += 1
            else:
                r += 1

        # Defaults value of the normalized taboo distance (Use for the subpopulations in the current restart)
        means_d = hat_d * np.ones((len(means),))
        sigmas = sigma * np.ones((len(means),))
        return means, means_d, sigmas, cov

    def new_iterate(self, means, means_d, sigmas, cov, archive_xx, archive_f, archive_d, args=None):
        # for termination criterion
        n_no_improvement = 120 + math.floor(30 * self.ndim_problem / self.n_individuals)  # No improvement
        n_stalled_size = n_no_improvement - 110  # Stalled size (stagnation)

        # initial values
        elt_xx = {k: np.tile(means[k], (self.n_elt, 1)) for k in range(self.n_s)}
        elt_zz = {k: 0 * np.tile(means[k], (self.n_elt, 1)) for k in range(self.n_s)}
        elt_f = {k: np.ones((self.n_elt,)) * self.best_so_far_y for k in range(self.n_s)}
        elt_s = {k: np.ones((self.n_elt,)) * np.mean(sigmas) for k in range(self.n_s)}
        cov_dict = {k: cov for k in range(self.n_s)}
        best_f_ne = {k: np.arange(n_no_improvement)[::-1] for k in range(self.n_s)}
        median_f_ne = {k: np.arange(n_no_improvement)[::-1] for k in range(self.n_s)}
        superior_index = {k: np.arange(k) for k in range(self.n_s)}
        best_xx = np.copy(means)
        best_f = np.ones((self.n_s,)) * self.best_so_far_y_bak

        active_subpop = np.arange(self.n_s)  # Activating subpopulations
        terminate_subpop = np.empty((0,), dtype=int)  # Terminated subpopulations
        used_iteration = np.ones((self.n_s,))
        terminate_flag, n_iteration = 0, 0
        while (active_subpop.size >= 1) and (terminate_flag == 0):
            for m in active_subpop:
                # The number of potential global minima in the current restart. It is used only for termination of the
                # restart so that sufficient budget remains for analyzing the subpopulations later.
                # This is a specialization for GECCO2016 competition
                cnt = np.ones((len(best_f),))[(best_f - self.fitness_threshold) <= min(np.hstack((archive_f, best_f)))]
                cnt = 1 + np.sum(cnt)
                if (self.max_function_evaluations - self.n_function_evaluations) <= \
                        (cnt * self.max_budget * (len(archive_xx) + cnt / 2) + self.n_individuals):
                    self.n_function_evaluations += (self.max_function_evaluations - self.n_function_evaluations) + 1
                    terminate_flag = 1  # terminate the restart
                    break

                # Termination condition
                condition_number = np.linalg.cond(cov_dict[m])
                max_diff = max(best_f_ne[m][-n_stalled_size:]) - min(best_f_ne[m][-n_stalled_size:])
                # the median of the 20 newest values is not smaller than the median of the 20 oldest values,
                # excluding elites
                min_imp_best = np.median(best_f_ne[m][-20:])
                min_imp_best = min_imp_best - np.median(best_f_ne[m][-n_no_improvement:(-n_no_improvement + 20)])
                min_imp_med = np.median(median_f_ne[m][-20:])
                min_imp_med = min_imp_med - np.median(median_f_ne[m][-n_no_improvement:(-n_no_improvement + 20)])
                active_flags = [condition_number < self.n_condition, max_diff > self.fitness_threshold,
                                min_imp_best <= 0, min_imp_med <= 0, not self._check_terminations()]
                if not all(active_flags):  # Terminate the subpopulation if a stopping criterion is satisfied
                    means[m] = best_xx[m]
                    terminate_subpop = np.append(terminate_subpop, m)
                    continue

                used_iteration[m] = n_iteration + 1
                superior_mean = means[superior_index[m]]
                superior_d = means_d[superior_index[m]]
                best_xx_f, new_mean_cov_sigma, elt_xx_f_s_zz, best_median_f_ne = \
                    self.generate_subpopulation(
                        means[m], cov_dict[m], sigmas[m],
                        superior_mean, superior_d,
                        archive_xx, archive_f, archive_d,
                        elt_xx[m], elt_f[m], elt_zz[m], elt_s[m],
                        best_f_ne[m], median_f_ne[m], args)
                best_xx[m], best_f[m] = best_xx_f
                means[m], cov_dict[m], sigmas[m] = new_mean_cov_sigma
                elt_xx[m], elt_f[m], elt_s[m], elt_zz[m] = elt_xx_f_s_zz
                best_f_ne[m], median_f_ne[m] = best_median_f_ne

                # Find non-terminated subpopulations
            active_subpop = np.array([i for i in active_subpop if i not in terminate_subpop])
            if active_subpop.size != 0:
                active_subpop = active_subpop[np.argsort(best_f[active_subpop])]
                superior_index.update({m: np.arange(len(best_f))[best_f < best_f[m]] for m in active_subpop})
            n_iteration += 1

            # Update parameters (Stagnation:)
            n_no_improvement = 120 + math.floor(30 * self.ndim_problem / self.n_individuals + 0.2 * n_iteration)

            if self._check_terminations():
                break

        self.average_iteration_subpopulations = np.mean(used_iteration)
        return best_xx, best_f

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        ES.optimize(self, fitness_function)
        archive_xx, archive_f, archive_d = np.empty((0, self.ndim_problem)), np.empty((0,)), np.empty((0,))
        all_best_xx, all_best_f = {}, {}
        while True:
            # Initialize subpopulations (Algorithm 2)
            means, means_d, sigmas, cov = self.initialize_subpopulations(archive_xx, archive_d)

            # Iterate
            best_xx, best_f = self.new_iterate(means, means_d, sigmas, cov, archive_xx, archive_f, archive_d, args)
            all_best_xx.update({self._n_generations + 1: best_xx})
            all_best_f.update({self._n_generations + 1: best_f})

            # Update the archive using the best solutions of the converged subpopulations and the old archive
            # (Algorithm 1)
            archive_xx, archive_f, archive_d = self.updating_archive(best_xx, best_f, archive_xx, archive_f, archive_d)

            # Restart (Equations 8)
            self.update_population_size()

            if self._check_terminations():
                break
            self._n_generations += 1
        # return archive_xx, archive_f, archive_d, all_best_xx, all_best_f
        return archive_f



import numpy as np

from pypop7.optimizers.ds.ds import DS


class GPS(DS):
    """Generalized Pattern Search (GPS).

    NOTE that "to converge to a local minimum, certain conditions must be met. The set of directions must
        be a positive spanning set, which means that we can construct any point using a nonnegative
        linear combination of the directions. A positive spanning set ensures that at least one of the
        directions is a descent direction from a location with a nonzero gradient."
        (from [Kochenderfer&Wheeler, 2019])

    Reference
    ---------
    Kochenderfer, M.J. and Wheeler, T.A., 2019.
    Algorithms for optimization.
    MIT Press.
    https://algorithmsbook.com/optimization/files/chapter-7.pdf
    (See Algorithm 7.6 (Page 106) for details.)

    Regis, R.G., 2016.
    On the properties of positive spanning sets and positive bases.
    Optimization and Engineering, 17(1), pp.229-262.
    https://link.springer.com/article/10.1007/s11081-015-9286-x

    Torczon, V., 1997.
    On the convergence of pattern search algorithms.
    SIAM Journal on Optimization, 7(1), pp.1-25.
    https://epubs.siam.org/doi/abs/10.1137/S1052623493250780
    """
    def __init__(self, problem, options):
        DS.__init__(self, problem, options)
        self.gamma = options.get('gamma', 0.5)  # decreasing factor of step-size (γ)

    def initialize(self, args=None, is_restart=False):
        x = self._initialize_x(is_restart)  # initial point
        y = self._evaluate_fitness(x, args)  # fitness
        # random directions
        d = self.rng_initialization.standard_normal(size=(self.ndim_problem + 1, self.ndim_problem))
        i_d = [i for i in range(d.shape[0])]  # index of used directions
        return x, y, d, i_d

    def iterate(self, args=None, x=None, d=None, i_d=None, fitness=None):
        improved, best_so_far_y = False, self.best_so_far_y
        for i in range(d.shape[0]):
            if self._check_terminations():
                return i_d
            x = self.best_so_far_x + self.sigma * d[i_d[i]]  # opportunistic
            y = self._evaluate_fitness(x, args)
            if self.record_fitness:
                fitness.append(y)
            if y < best_so_far_y:
                improved = True
                i_d = [i_d[i]] + i_d[:i] + i_d[(i + 1):]  # dynamic ordering
                break
        if not improved:
            self.sigma *= self.gamma  # alpha
        return i_d

    def restart_initialize(self, args=None, x=None, y=None, d=None, i_d=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            x, y, d, i_d = self.initialize(args, is_restart)
            fitness.append(y)
            self._fitness_list = [self.best_so_far_y]
        return x, y, d, i_d

    def optimize(self, fitness_function=None, args=None):
        fitness = DS.optimize(self, fitness_function)
        x, y, d, i_d = self.initialize(args)
        fitness.append(y)
        while True:
            i_d = self.iterate(args, x, d, i_d, fitness)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, y, d, i_d = self.restart_initialize(args, x, y, d, i_d, fitness)
        results = self._collect_results(fitness)
        return results



import numpy as np

from optimizers.pso.pso import PSO


# helper function
def global_topology(p_x, p_y, i):
    return p_x[np.argmin(p_y)], i


class PSOGT(PSO):
    """Particle Swarm Optimizer with Global Topology (PSOGT).

    Reference
    ---------
    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (Vol. 4, pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146
    """
    def __init__(self, problems, options):
        PSO.__init__(self, problems, options)
        self.topology = global_topology



"""Repeat Fig.3, Fig.4 from the following paper:
    I. Loshchilov
    LM-CMA: an Alternative to L-BFGS for Large Scale Black-box Optimization
    Evolutionary Computation 2017; 25(1): 143-171
    https://direct.mit.edu/evco/article-abstract/25/1/143/1041/LM-CMA-An-Alternative-to-L-BFGS-for-Large-Scale?redirectedFrom=fulltext
"""
import sys
import pickle
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock, discus, cigar, different_powers
from pypop7.benchmarks.rotated_functions import generate_rotation_matrix
from pypop7.benchmarks.rotated_functions import ellipsoid as rotated_ellipsoid
from pypop7.benchmarks.rotated_functions import rosenbrock as rotated_rosenbrock
from pypop7.benchmarks.rotated_functions import discus as rotated_discus
from pypop7.benchmarks.rotated_functions import cigar as rotated_cigar
from pypop7.benchmarks.rotated_functions import different_powers as rotated_different_powers
from pypop7.optimizers.core import optimizer
from pypop7.optimizers.es.lmcma import LMCMA as Solver

sys.modules['optimizer'] = optimizer
sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function, problem_dim):
    plt.figure()
    result = read_pickle(function, problem_dim)
    print(result)
    plt.yscale('log')
    plt.xscale('log')
    if function in ['ellipsoid', 'rotated_ellipsoid']:
        plt.ylim([1e-10, 1e10])
        plt.yticks([1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2, 1e4, 1e6, 1e8, 1e10])
        plt.xlim([1e2, 2e8])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
    elif function in ['rosenbrock', 'rotated_rosenbrock']:
        plt.ylim([1e-10, 1e9])
        plt.yticks([1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2, 1e4, 1e6, 1e8])
        plt.xlim([1e2, 1e9])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9])
    elif function in ['discus', 'rotated_discus']:
        plt.ylim([1e-10, 1e8])
        plt.yticks([1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2, 1e4, 1e6, 1e8])
        plt.xlim([1e0, 1e8])
        plt.xticks([1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
    elif function in ['cigar', 'rotated_cigar']:
        plt.ylim([1e-10, 1e12])
        plt.yticks([1e-10, 1e-5, 1e0, 1e5, 1e10])
        plt.xlim([1e2, 1e7])
        plt.xticks([1e2, 1e3, 1e4, 1e5, 1e6])
    elif function in ['different_powers', 'rotated_different_powers']:
        plt.ylim([1e-10, 1e7])
        plt.yticks([1e-10, 1e-8, 1e-6, 1e-4, 1e-2, 1e0, 1e2, 1e4, 1e6])
        plt.xlim([1e0, 1e8])
        plt.xticks([1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='blue')
    plt.xlabel("Function Evaluations")
    plt.ylabel("Objective Function")
    plt.title(function.capitalize())
    plt.savefig('lmcma/result_' + function + '.png')
    plt.show()


if __name__ == '__main__':
    for f in [ellipsoid, rosenbrock, discus, cigar, different_powers, rotated_ellipsoid,
              rotated_rosenbrock, rotated_discus, rotated_cigar, rotated_different_powers]:
        for d in [512]:
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'max_function_evaluations': 2e7,
                       'max_runtime': 3600,
                       'fitness_threshold': 1e-10,
                       'seed_rng': 0,  # undefined in the original paper
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 3.0,
                       'verbose_frequency': 2e4,
                       'record_fitness': True,
                       'record_fitness_frequency': 1,
                       'is_restart': False}
            if f in [rotated_ellipsoid, rotated_rosenbrock, rotated_discus,
                              rotated_cigar, rotated_different_powers]:
                generate_rotation_matrix(f, d, 2022)
                function_name = 'rotated_' + f.__name__
            else:
                function_name = f.__name__
            solver = Solver(problem, options)
            results = solver.optimize()
            write_pickle(function_name, d, results)
            plot(function_name, d)



import numpy as np

from optimizers.es.es import ES


class NES(ES):
    """Natural Evolution Strategies (NES).

    Reference
    ---------
    Wierstra, D., Schaul, T., Glasmachers, T., Sun, Y., Peters, J. and Schmidhuber, J., 2014.
    Natural evolution strategies.
    Journal of Machine Learning Research, 15(27), pp.949-980.
    https://jmlr.org/papers/v15/wierstra14a.html
    """
    def __init__(self, problem, options):
        problem['_is_maximization'] = True  # mandatory setting for NES
        ES.__init__(self, problem, options)

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def optimize(self, fitness_function=None):
        raise NotImplementedError

    def _compute_gradients(self):
        raise NotImplementedError

    def _update_distribution(self):
        raise NotImplementedError

    def _fitness_shaping(self):
        base = np.log(self.n_individuals / 2 + 1)
        utilities = np.maximum(0, [base - np.log(k) for k in (np.arange(self.n_individuals) + 1)])
        return utilities / np.sum(utilities) - (1 / self.n_individuals)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class CC(Optimizer):
    """Cooperative Coevolution
    Reference
    -----------
    F. Gomez, J. Schmidhuber, R. Miikkulainen
    Accelerated Neural Evolution through Cooperatively Coevolved Synapses
    https://jmlr.org/papers/v9/gomez08a.html
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of offspring, offspring population size (λ: lambda)
            self.n_individuals = 4 + int(3 * np.log(self.ndim_problem))  # for small population setting
        if self.n_parents is None:  # number of parents, parental population size (μ: mu)
            self.n_parents = int(self.n_individuals / 2)
        self._n_generations = 0
        self.x = options.get('x')

    def initialize(self, is_restart=False):
        raise NotImplementedError

    def iterate(self, mean, x, y):
        raise NotImplementedError

    def crossover(self, x, y, cross_type):
        s1 = x.copy()
        s2 = y.copy()
        if cross_type == 'one_point':
            place = np.random.randint(0, self.ndim_problem)
            for i in range(place):
                s1[i] = y[i]
                s2[i] = x[i]
        elif cross_type == 'two_point':
            place1 = np.random.randint(0, self.ndim_problem)
            place2 = np.random.randint(place1, self.ndim_problem)
            for i in range(place1, place2):
                s1[i] = y[i]
                s2[i] = x[i]
        elif cross_type == 'uniform':
            for i in range(self.ndim_problem):
                rand = np.random.random()
                if rand < 0.5:
                    s1[i] = y[i]
                    s2[i] = x[i]
        return [s1, s2]

    def mutate(self, x):
        for i in range(self.ndim_problem):
            rand = np.random.random()
            if rand < self.prob_mutate:
                x[i] = self.lower_boundary[i] + np.random.random() *\
                       (self.upper_boundary[i] - self.lower_boundary[i])
        return x

    def _initialize_x(self, is_restart=False):
        if is_restart or (self.x is None):
            x = self.rng_initialization.uniform(self.initial_lower_boundary,
                                                self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        return x

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            best_so_far_y = -self.best_so_far_y if self._is_maximization else self.best_so_far_y
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness, mean=None):
        results = Optimizer._collect_results(self, fitness)
        results['sigma'] = self.sigma
        results['_n_generations'] = self._n_generations
        return results



"""Repeat Figure 1 of the JMLR-2014-NES paper:
    Wierstra, D., Schaul, T., Glasmachers, T., Sun, Y., Peters, J. and Schmidhuber, J., 2014.
    Natural evolution strategies.
    Journal of Machine Learning Research, 15(1), pp.949-980.
    https://jmlr.org/papers/volume15/wierstra14a/wierstra14a.pdf
"""

import numpy as np
from numpy.random import default_rng
import matplotlib.pyplot as plt


class CanonicalSearchGradientAlgorithm(object):
    """Canonical Search Gradient Algorithm Supporting Only One-Dimensional Gaussian Distribution.
        See Algorithm 1, Algorithm 2, Section 2.2, and Figure 1 of the JMLR-2014-NES paper for more details.
    """
    def __init__(self, f, ndim_problem=1, max_generations=1e3, eta=0.01, n_individuals=10, seed_rng=0):
        self.f = f  # fitness function (problem) to be minimized
        self.ndim_problem = ndim_problem  # number of dimensionality of problem
        self.max_generations = max_generations  # maximum of generations
        self.eta = eta  # learning rate
        self.n_individuals = n_individuals  # number of individuals (population size, lambda)
        self.seed_rng = seed_rng  # seed of random number generator (rng)

    def initialize(self, mu=1.0, sigma=1.0):
        z = np.empty((self.n_individuals, self.ndim_problem))  # population
        f_z = np.empty((self.n_individuals,))  # fitness of population
        g_mu = np.empty((self.n_individuals, self.ndim_problem))  # log-derivatives of mean of Gaussian distribution
        g_sigma = np.empty((self.n_individuals, self.ndim_problem))  # log-derivatives of std of Gaussian distribution
        history = {"mu": [mu], "sigma": [sigma]}
        return z, f_z, g_mu, g_sigma, mu, sigma, history

    def iterate(self, z, f_z, g_mu, g_sigma, mu, sigma):
        inv_sigma = np.linalg.inv(np.array([[sigma]]))
        for k in range(self.n_individuals):
            # draw sample
            z[k] = mu + sigma * default_rng(self.seed_rng).standard_normal((self.ndim_problem,))
            # evaluate fitness
            f_z[k] = self.f(z[k])
            # calculate fitness-weighted log-derivatives
            zk_minus_mu = z[k] - mu
            g_mu[k] = (inv_sigma * zk_minus_mu) * f_z[k]
            g_sigma[k] = (-0.5 * inv_sigma + 0.5 * inv_sigma * zk_minus_mu * zk_minus_mu * inv_sigma) * f_z[k]
        return g_mu, g_sigma

    def optimize(self):
        z, f_z, g_mu, g_sigma, mu, sigma, history = self.initialize()
        for _ in range(int(self.max_generations)):
            g_mu, g_sigma = self.iterate(z, f_z, g_mu, g_sigma, mu, sigma)
            mu -= float(self.eta * np.mean(g_mu, 0))
            sigma -= float(self.eta * np.mean(g_sigma, 0))
            history["mu"].append(mu)
            history["sigma"].append(sigma)
        return history


if __name__ == "__main__":
    def sphere(x):
        return np.power(x, 2)
    n_trials = 3
    stat_mean, stat_std = np.empty((n_trials, 1001)), np.empty((n_trials, 1001))
    for t in range(n_trials):
        sg = CanonicalSearchGradientAlgorithm(sphere, seed_rng=t)
        history_mean_std = sg.optimize()
        stat_mean[t] = np.array(history_mean_std["mu"])
        stat_std[t] = np.array(history_mean_std["sigma"])
    plt.plot(np.median(stat_mean, 0), "black")
    plt.plot(np.median(stat_std, 0), "red")
    plt.yscale("log")
    plt.show()



"""Repeat Fig.4, Fig.5 in this paper
    (1+1)-Cholesky-CMA-ES (OPOC).
    Reference
    ---------
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 2 for details.)
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle

from pypop7.optimizers.es.opoc2009 import OPOC2009 as Solver
from pypop7.benchmarks.base_functions import ellipsoid, cigar

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


# plot fig4
def plot(function, problem_dim):
    plt.figure()
    result = read_pickle(function, problem_dim)
    plt.yscale('log')
    plt.ylim([1e-250, 1e50])
    plt.yticks([1e-250, 1e-200, 1e-150, 1e-100, 1e-50, 1e0, 1e50])
    plt.xticks([0, 2e4, 4e4, 6e4, 8e4, 1e5])
    plt.xlim([0, 1e5])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='green')
    plt.xlabel("objective function evaluations")
    plt.ylabel("objective function value")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    # fig.4
    for f in [ellipsoid]:
        for d in [20]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'fitness_threshold': 1e-250,
                       'max_runtime': 3600,  # 1 hours
                       'seed_rng': 0,
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 3.0,
                       'max_function_evaluations': 6e4,
                       'verbose_frequency': 2000,
                       'record_fitness': True,
                       'record_fitness_frequency': 1,
                       'is_restart': False}
            solver = Solver(problem, options)
            results = solver.optimize()
            write_pickle(f.__name__, d, results)
            plot(f.__name__, d)
    # fig 5
    evaluations = []
    ds = [5, 10, 15, 20, 40, 80, 160, 320]
    for f in [cigar]:
        for d in ds:
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'fitness_threshold': 1e-15,
                       'max_runtime': 3600,  # 1 hours
                       'seed_rng': 0,
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 3.0,
                       'max_function_evaluations': 2e5,
                       'verbose_frequency': 2000,
                       'record_fitness': False,
                       'record_fitness_frequency': 1,
                       'is_restart': False}
            solver = Solver(problem, options)
            results = solver.optimize()
            evaluations.append(results['n_function_evaluations'])
    plt.figure()
    plt.yscale('log')
    plt.xscale('log')
    plt.ylim([1e3, 1e7])
    plt.yticks([1e3, 1e4, 1e5, 1e6, 1e7])
    plt.xlim([1, 4e2])
    plt.xticks([10, 100])
    plt.plot(ds, evaluations, 'ro--', color='green')
    plt.xlabel('search space dimension n')
    plt.ylabel('objective function evaluations')
    plt.show()



from pypop7.optimizers.es.maes import MAES


class FMAES(MAES):
    """Fast Matrix Adaptation Evolution Strategy (FMAES, Fast-(μ/μ_w,λ)-MA-ES).

    Reference
    ---------
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700).
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    Loshchilov, I., Glasmachers, T. and Beyer, H.G., 2019.
    Large scale black-box optimization by limited-memory matrix adaptation.
    IEEE Transactions on Evolutionary Computation, 23(2), pp.353-358.
    https://ieeexplore.ieee.org/abstract/document/8410043

    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115

    See the official Matlab version from Beyer:
    https://homepages.fhv.at/hgb/downloads/ForDistributionFastMAES.tar
    """
    def __init__(self, problem, options):
        options['_fast_version'] = True  # mandatory setting for FMAES
        MAES.__init__(self, problem, options)



import numpy as np

from pypop7.optimizers.cc.cc import CC


class COSYNE(CC):
    """Cooperative Synapse Neuroevolution(CoSyNE)
    Reference
    -----------
    F. Gomez, J. Schmidhuber, R. Miikkulainen
    Accelerated Neural Evolution through Cooperatively Coevolved Synapses
    https://jmlr.org/papers/v9/gomez08a.html
    """
    def __init__(self, problem, options):
        CC.__init__(self, problem, options)
        self.n_combine = options.get('n_combine')
        self.crossover_type = options.get('crossover_type')
        self.prob_mutate = options.get('prob_mutate')

    def initialize(self, is_restart=False):
        x = np.empty((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            x[i] = self._initialize_x()
        return x, y

    def mutate(self, x):
        for i in range(2):
            for j in range(self.ndim_problem):
                rand = np.random.random()
                if rand < self.prob_mutate:
                    x[i][j] = x[i][j] + 0.3 * self.rng_optimization.standard_cauchy(1)
            x[i] = np.clip(x[i], self.lower_boundary, self.upper_boundary)
        return x

    def permute(self, log):
        temp = []
        length = len(log)
        for k in range(length):
            index = np.random.randint(0, len(log)) % len(log)
            temp.append(log[index])
            log.pop(index)
        return temp

    def iterate(self, x, y):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            y[i] = self._evaluate_fitness(x[i])
        order = np.argsort(y)
        o = self.crossover(x[order[0]], x[order[1]], self.crossover_type)
        o = self.mutate(o)
        for i in range(self.ndim_problem):
            p = x[:, i]
            order_1 = np.argsort(p)
            for k in range(2):
                x[order_1[self.n_individuals - k - 1]][i] = o[k][i]
            mark_weight, log = [], []
            for j in range(self.n_individuals):
                temp = (y[j] - y[order[0]])/(y[order[-1]] - y[order[0]])
                prob = np.power(temp, 1.0 / self.ndim_problem)
                rand = np.random.random()
                if rand < prob:
                    mark_weight.append(x[j][i])
                    log.append(j)
            log = self.permute(log)
            for j in range(len(log)):
                x[log[j]][i] = mark_weight[j]
        return x, y

    def optimize(self, fitness_function=None):
        fitness = CC.optimize(self, fitness_function)
        x, y = self.initialize()
        while True:
            x, y = self.iterate(x, y)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self.__collect_results(fitness)
        return results



"""Repeat Fig.3(Right) from the following paper:
     Loshchilov, I., 2014, July.
    A computationally efficient limited memory CMA-ES for large scale optimization.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 397-404). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598294
"""
import sys
import pickle
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.rotated_functions import ellipsoid, generate_rotation_matrix
from pypop7.optimizers.core import optimizer
from pypop7.optimizers.es.lmcmaes import LMCMAES as Solver

sys.modules['optimizer'] = optimizer
sns.set_theme(style='darkgrid')


def read_pickle(function):
    file = function + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, result):
    file = open(function + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function):
    plt.figure()
    result = read_pickle(function)
    result = np.array(result)
    plt.yscale('log')
    plt.xscale('log')
    plt.ylim([1e4, 1e10])
    plt.yticks([1e4, 1e5, 1e6, 1e7, 1e8, 1e9, 1e10])
    plt.xlim([1e2, 1e4])
    plt.xticks([1e2, 1e3, 1e4])
    plt.plot(result[:, 0], result[:, 1], color='purple', marker='s')
    plt.xlabel("Dimension")
    plt.ylabel("Function Evaluations")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    function_evaluations = []
    for f in [ellipsoid]:
        for d in [128, 256, 512, 1024, 2048, 4096, 8192]:
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'max_function_evaluations': 1e10,
                       'fitness_threshold': 1e-10,
                       'seed_rng': 0,  # undefined in the original paper
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 5,
                       'verbose_frequency': 5000,
                       'record_fitness': False,
                       'record_fitness_frequency': 1,
                       'is_restart': False}
            generate_rotation_matrix(f, d, 2022)
            solver = Solver(problem, options)
            results = solver.optimize()
            function_evaluations.append([d, results['n_function_evaluations']])
        write_pickle(f.__name__, function_evaluations)
        plot(f.__name__)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class CODE(Optimizer):
    """Composite Differential Evolution (CoDE).

    Reference
    ---------
    Wang, Y., Cai, Z., and Zhang, Q. 2011.
    Differential evolution with composite trial vector generation strategies and control parameters.
    IEEE Transactions on Evolutionary Computation, 15(1), 55–66.
    https://doi.org/10.1109/TEVC.2010.2087271
    """

    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        self._n_generations = 0
        self.c_p = [[1.0, 0.1], [1.0, 0.9], [0.8, 0.2]]  # [f, cr] control parameter settings pool
        self.n_individuals = options.get('n_individuals', 30)

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(
            self.initial_lower_boundary, self.initial_upper_boundary,
            (self.n_individuals, self.ndim_problem))  # initial point
        y = []
        for i in range(self.n_individuals):
            y.append(self._evaluate_fitness(x[i], args))
        return x, y

    def mutation(self, x, xx=None):
        # Mutation
        c = self.rng_optimization.choice([0, 1, 2], (3,))  # randomly selected from the parameter candidate pool
        p = self.rng_optimization.permutation(self.n_individuals)
        v_1 = xx[p[0]] + self.c_p[c[0]][0] * (xx[p[1]] - xx[p[2]])  # rand/1/bin

        p = self.rng_optimization.permutation(self.n_individuals)
        v_2 = xx[p[0]] + self.c_p[c[1]][0] * (xx[p[1]] - xx[p[2]]) + \
              self.c_p[c[1]][0] * (xx[p[3]] - xx[p[4]])  # rand/2/bin

        p = self.rng_optimization.permutation(self.n_individuals)
        v_3 = x + self.rng_optimization.random() * (xx[p[0]] - x) + \
              self.c_p[c[2]][0] * (xx[p[1]] - xx[p[2]])  # current-to-rand/1
        return [v_1, v_2, v_3], c

    def crossover(self, v, x, cr):
        u = np.zeros((self.ndim_problem,))
        for j in range(self.ndim_problem):
            if (j == self.rng_optimization.integers(0, self.ndim_problem)) or \
                    (self.rng_optimization.random() < cr):
                u[j] = v[j]
            else:
                u[j] = x[j]
        return u

    def iterate(self, args=None, x=None, fitness=None):
        u = np.zeros((self.n_individuals, self.ndim_problem))
        y_u = []
        for i in range(self.n_individuals):
            # mutation
            v, c = self.mutation(x[i], x)
            for j in [0, 1]:
                v[j] = self.crossover(v[j], x[i], self.c_p[c[j]][1])

            for j in [0, 1, 2]:
                b = v[j] < self.lower_boundary  # bool values
                v[j][b] = np.minimum(self.upper_boundary, 2 * self.lower_boundary - v[j])[b]
                b = v[j] > self.upper_boundary  # bool values
                v[j][b] = np.maximum(self.lower_boundary, 2 * self.upper_boundary - v[j])[b]

            # choose the best trial vector
            y_v = [self._evaluate_fitness(v[j], args) for j in range(len(v))]
            idx = np.argsort(y_v)[0]
            u[i] = v[idx]
            y_u.append(y_v[idx])

        # Selection
        x_new = np.copy(x)
        y_new = fitness[-self.n_individuals:]
        b = y_u < y_new  # bool values
        x_new[b] = u[b]
        y_new[b] = y_u[b]
        fitness.extend(y_new)
        return x_new, fitness

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        x, y = self.initialize(args)
        fitness.extend(y)
        while True:
            x, fitness = self.iterate(args, x, fitness)

            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
        results = self._collect_results(fitness)
        return results



import numpy as np

from pypop7.optimizers.es.es import ES


class R1ES(ES):
    """Rank-One Evolution Strategy (R1ES).

    Reference
    ---------
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_cov = options.get('c_cov', 1 / (3 * np.sqrt(self.ndim_problem) + 5))  # for Line 5 in Algorithm 1
        self.c = options.get('c', 2 / (self.ndim_problem + 7))  # for Line 12 in Algorithm 1 (c_c)
        self.c_s = options.get('c_s', 0.3)  # for Line 15 in Algorithm 1
        self.q_star = options.get('q_star', 0.3)  # for Line 15 in Algorithm 1
        self.d_sigma = options.get('d_sigma', 1)  # for Line 16 in Algorithm 1
        self._x_1 = np.sqrt(1 - self.c_cov)  # for Line 5 in Algorithm 1
        self._x_2 = np.sqrt(self.c_cov)  # for Line 5 in Algorithm 1
        self._p_1 = 1 - self.c  # for Line 12 in Algorithm 1
        self._p_2 = None  # for Line 12 in Algorithm 1
        self._rr = None  # for rank-based success rule (RSR)

    def initialize(self, args=None, is_restart=False):
        self._p_2 = np.sqrt(self.c * (2 - self.c) * self._mu_eff)
        self._rr = np.arange(self.n_parents * 2) + 1
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        p = np.zeros((self.ndim_problem,))  # principal search direction
        s = 0  # cumulative rank rate
        y = np.tile(self._evaluate_fitness(mean, args), (self.n_individuals,))  # fitness
        return x, mean, p, s, y

    def iterate(self, x=None, mean=None, p=None, y=None, args=None):
        for k in range(self.n_individuals):  # for Line 3 in Algorithm 1
            if self._check_terminations():
                return x, y
            z = self.rng_optimization.standard_normal((self.ndim_problem,))  # for Line 4 in Algorithm 1
            r = self.rng_optimization.standard_normal()  # for Line 4 in Algorithm 1
            # for Line 5 in Algorithm 1
            x[k] = mean + self.sigma * (self._x_1 * z + self._x_2 * r * p)
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, p=None, s=None, y=None, y_bak=None):
        order = np.argsort(y)
        y.sort()  # for Line 10 in Algorithm 1
        # for Line 11 in Algorithm 1
        mean_w = np.zeros((self.ndim_problem,))
        for k in range(self.n_parents):
            mean_w += self._w[k] * x[order[k]]
        p = self._p_1 * p + self._p_2 * (mean_w - mean) / self.sigma  # for Line 12 in Algorithm 1
        mean = mean_w  # for Line 11 in Algorithm 1
        # for rank-based adaptation of mutation strength
        r = np.argsort(np.hstack((y_bak[:self.n_parents], y[:self.n_parents])))
        rr = self._rr[r < self.n_parents] - self._rr[r >= self.n_parents]
        q = np.dot(self._w, rr) / self.n_parents  # for Line 14 in Algorithm 1
        s = (1 - self.c_s) * s + self.c_s * (q - self.q_star)  # for Line 15 in Algorithm 1
        self.sigma *= np.exp(s / self.d_sigma)  # for Line 16 in Algorithm 1
        return mean, p, s

    def restart_initialize(self, args=None, x=None, mean=None, p=None, s=None, y=None, fitness=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            x, mean, p, s, y = self.initialize(args, is_restart)
            fitness.append(y[0])
            self.d_sigma *= 2
        return x, mean, p, s, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p, s, y = self.initialize(args)
        fitness.append(y[0])
        while True:
            y_bak = np.copy(y)  # for Line 13 in Algorithm 1
            x, y = self.iterate(x, mean, p, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p, s = self._update_distribution(x, mean, p, s, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, mean, p, s, y = self.restart_initialize(args, x, mean, p, s, y, fitness)
        results = self._collect_results(fitness, mean)
        results['p'] = p
        results['s'] = s
        return results



import numpy as np

from pypop7.optimizers.es.opoc2009 import OPOC2009


class OPOA(OPOC2009):
    """(1+1)-Active-CMA-ES (OPOA).

    Reference
    ---------
    Arnold, D.V. and Hansen, N., 2010, July.
    Active covariance matrix adaptation for the (1+1)-CMA-ES.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 385-392). ACM.
    https://dl.acm.org/doi/abs/10.1145/1830483.1830556
    """
    def __init__(self, problem, options):
        OPOC2009.__init__(self, problem, options)
        self.c_m = options.get('c_m', 0.4 / (np.power(self.ndim_problem, 1.6) + 1))
        self.k = options.get('k', 5)
        self._ancestors = []

    def initialize(self, args=None, is_restart=False):
        mean, y, a, a_i, best_so_far_y, p_s, p_c = OPOC2009.initialize(self, args, is_restart)
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def iterate(self, args=None, mean=None, a=None, a_i=None, best_so_far_y=None, p_s=None, p_c=None):
        # sample and evaluate (only one) offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        x = mean + self.sigma * np.dot(a, z)
        y = self._evaluate_fitness(x, args)
        if y <= best_so_far_y:
            l_s = 1
        else:
            l_s = 0
        p_s = (1 - self.c_p) * p_s + self.c_p * l_s
        self.sigma *= np.exp(self.eta_sigma * (p_s - self.p_ts) / (1 - self.p_ts))
        if y <= best_so_far_y:
            self._ancestors.append(y)
            mean, best_so_far_y = x, y
            if p_s < self.p_t:
                p_c = (1 - self.c_c) * p_c + np.sqrt(self.c_c * (2 - self.c_c)) * np.dot(a, z)
                alpha = 1 - self.c_cov
            else:
                p_c *= 1 - self.c_c
                alpha = 1 - self.c_cov + self.c_cov * self.c_c * (2 - self.c_c)
            w = np.dot(a_i, p_c)
            w_power = np.dot(w, w)
            alpha = np.sqrt(alpha)
            beta = alpha / w_power * (np.sqrt(1 + self.c_cov / (1 - self.c_cov) * w_power) - 1)
            a = alpha * a + beta * np.dot(p_c[:, np.newaxis], w[np.newaxis, :])
            a_i = 1 / alpha * a_i - beta / (np.power(alpha, 2) + alpha * beta * w_power) * np.dot(
                w[:, np.newaxis], np.dot(w[np.newaxis, :], a_i))
        if len(self._ancestors) >= self.k and y > self._ancestors[-self.k]:
            del self._ancestors[0]
            z_power = np.dot(z, z)
            if 1 < self.c_m * (2 * z_power - 1):
                c_m = 1 / (2 * z_power - 1)
            else:
                c_m = self.c_m
            alpha = np.sqrt(1 + c_m)
            beta = alpha / z_power * (np.sqrt(1 - self.c_m / (1 - self.c_m) * z_power) - 1)
            a = alpha * a + beta * np.dot(np.dot(a, z[:, np.newaxis]), z[np.newaxis, :])
            a_i = 1 / alpha * a_i - beta / (np.power(alpha, 2) + alpha * beta * z_power) * np.dot(
                z[:, np.newaxis], np.dot(z[np.newaxis, :], a_i))
        return mean, y, a, a_i, best_so_far_y, p_s, p_c

    def restart_initialize(self, args=None, mean=None, y=None, a=None, a_i=None,
                           best_so_far_y=None, p_s=None, p_c=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args, is_restart)
            fitness.append(y)
            self._fitness_list = [best_so_far_y]
            self._ancestors = []
        return mean, y, a, a_i, best_so_far_y, p_s, p_c



import unittest
import numpy as np
import time

from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock, rastrigin
from optimizers.pso.psort import PSORT


class TestPSORT(unittest.TestCase):
    def test_run(self):
        start_run = time.time()
        ndim_problem = 1000
        for f in [ellipsoid, rosenbrock, rastrigin]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': ndim_problem,
                       'lower_boundary': -5 * np.zeros((ndim_problem,)),
                       'upper_boundary': 5 * np.ones((ndim_problem,))}
            options = {'max_function_evaluations': 2e6,
                       'fitness_threshold': 1e-10,
                       'seed_rng': 0,
                       'x': 4 * np.ones((ndim_problem,)),
                       'verbose_options': {'frequency_verbose': 20000},
                       'record_options': {'record_fitness': True,
                                          'frequency_record_fitness': 200000}}
            psort = PSORT(problem, options)
            results = psort.optimize()
            print(results)
            print('*** Runtime: {:7.5e}'.format(time.time() - start_run))


if __name__ == '__main__':
    unittest.main()



import numpy as np

from pypop7.optimizers.cem.cem import CEM
import colorednoise

class SECEM(CEM):
    """Sample-efficient Cross-Entropy Method
    Reference
    ---------------
    C. Pinneri, S. Sawant, S. Blaes, J. Achterhold, J. Stuckler, M. Rolinek, G. Martius
    Sample-efficient Cross-Entropy Method for Real-time Planning
    https://arxiv.org/pdf/2008.06389.pdf
    """
    def __init__(self, problem, options):
        CEM.__init__(self, problem, options)
        self.alpha = 0.1
        self.fraction_reused_elites = options.get('fraction_reused_elites')
        self.decay = 1.25  # decay parameter used in N
        self.noise_beta = options.get('noise_beta')

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)
        y = np.empty((self.n_individuals,))
        x = np.empty((self.n_individuals, self.ndim_problem))
        elite_x = np.empty((self.n_parents, self.ndim_problem))
        elite_y = np.empty((self.n_parents,))
        return mean, x, elite_x, y, elite_y

    def re_initialize(self):
        y = np.empty((self.n_individuals,))
        return y

    def iterate(self, mean, x, y):
        for i in range(self.n_individuals):
            x[i] = colorednoise.powerlaw_psd_gaussian(self.noise_beta, size=(1, self.ndim_problem))
            x[i] = np.clip(np.dot(x[i], self.sigma) + self.mean, self.lower_boundary, self.upper_boundary)
            y[i] = self._evaluate_fitness(x[i])
            if self._check_terminations():
                return x, y
        return x, y

    def _update_parameters(self, mean, x, elite_x, y, elite_y):
        if self._n_generations == 0:
            order = np.argsort(y)
            for j in range(self.n_parents):
                elite_x[j] = x[order[j]]
                elite_y[j] = y[order[j]]
        else:
            length = int(self.fraction_reused_elites * self.n_parents)
            combine_x = np.empty((length, self.ndim_problem))
            combine_y = np.empty((length,))
            for k in range(length):
                combine_x[k] = elite_x[k]
                combine_y[k] = elite_y[k]
            combine_x = np.vstack((x, combine_x))
            combine_y = np.hstack((y, combine_y))
            order = np.argsort(combine_y)
            for j in range(self.n_parents):
                elite_x[j] = combine_x[order[j]]
                elite_y[j] = combine_y[order[j]]
        mean = self.alpha * mean + (1 - self.alpha) * np.mean(elite_x, axis=0)
        self.sigma = self.alpha * self.sigma + (1 - self.alpha) * np.std(elite_x, axis=0)
        return mean, elite_x, elite_y

    def optimize(self, fitness_function=None, args=None):
        fitness = CEM.optimize(self, fitness_function)
        mean, x, elite_x, y, elite_y = self.initialize()
        restart = True
        while True:
            self.n_individuals = int(max(self.n_individuals * np.power(self.decay, -1 * self._n_generations),
                                     2 * self.n_parents))
            if self.n_individuals != 2 * self.n_parents or restart is True:
                y = self.re_initialize()
                if self.n_individuals == 2 * self.n_parents:
                    restart = False
            x, y = self.iterate(mean, x, y)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, elite_x, elite_y = self._update_parameters(mean, x, elite_x, y, elite_y)
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness, mean)
        return results



from optimizers.pso.pso import PSO


# helper function
def ring_topology(p_x, p_y, i):
    if len(p_y) < 3:
        raise ValueError('Swarm size should (at least) >= 3.')
    if i == 0:
        left, right = len(p_y) - 1, 1
    elif i == len(p_y) - 1:
        left, right = len(p_y) - 2, 0
    else:
        left, right = i - 1, i + 1
    better = p_x[left] if p_y[left] <= p_y[right] else p_x[right]
    return better, i


class PSORT(PSO):
    """Particle Swarm Optimizer with Ring Topology (PSORT).

    Reference
    ---------
    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (Vol. 4, pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146
    """
    def __init__(self, problems, options):
        PSO.__init__(self, problems, options)
        self.topology = ring_topology



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class ES(Optimizer):
    """Evolution Strategies (ES).

    This is the **base** (abstract) class for all ES classes. Please use any of its concrete subclasses to
    optimize the black-box problem at hand.

    For interesting applications of ES, see e.g. `<>`_ for fluid dynamics and structural engineering.

    .. note:: Its three methods (`initialize`, `iterate`, `optimize`) should be implemented by its subclasses.

       `ES` are a well-established family of randomized population-based search algorithms, proposed originally by two
       German computer scientists Rechenberg and Schwefel (recipients of `Evolutionary Computation Pioneer Award 2002
       <https://tinyurl.com/456as566>`_). One key property of `ES` lies in its adaptability of strategy parameters,
       which can *significantly* accelerate the (local) convergence rate in many cases. Recently, the theoretical
       foundation of its most representative (modern) version called **CMA-ES** has been built on the very interesting
       `Information-Geometric Optimization (IGO) <https://www.jmlr.org/papers/v18/14-467.html>`_ framework via
       invariance principles.

       According to the latest `Nature <https://www.nature.com/articles/nature14544.>`_ review, *"the CMA-ES algorithm
       is widely regarded as the state of the art in numerical optimization"*.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
                * 'record_fitness'           - flag to record fitness list to output results (`bool`, default: `False`),
                * 'record_fitness_frequency' - function evaluations frequency of recording (`int`, default: `1000`),

                  * if `record_fitness` is set to `False`, it will be ignored,
                  * if `record_fitness` is set to `True` and it is set to 1, all fitness generated during optimization
                    will be saved into output results.

                * 'verbose'                  - flag to print verbose info during optimization (`bool`, default: `True`),
                * 'verbose_frequency'        - frequency of printing verbose info (`int`, default: `10`);
              and with four particular settings (`keys`):
                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`),
                * 'n_parents'     - number of parents (μ: mu), parental population size (`int`),
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`).

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring/descendants (λ: lambda), offspring population size.
    n_parents     : `int`
                    number of parents (μ: mu), parental population size.
    mean          : `array_like`
                    initial (starting) point, mean of Gaussian search distribution.
    sigma         : `float`
                    initial global step-size (σ), mutation strength (rate).

    Methods
    -------

    References
    ----------
    Ollivier, Y., Arnold, L., Auger, A. and Hansen, N., 2017.
    Information-geometric optimization algorithms: A unifying picture via invariance principles.
    Journal of Machine Learning Research, 18(18), pp.1-65.
    https://www.jmlr.org/papers/v18/14-467.html

    https://blog.otoro.net/2017/10/29/visual-evolution-strategies/

    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44

    http://www.scholarpedia.org/article/Evolution_strategies

    Beyer, H.G. and Schwefel, H.P., 2002.
    Evolution strategies–A comprehensive introduction.
    Natural Computing, 1(1), pp.3-52.
    https://link.springer.com/article/10.1023/A:1015059928466

    Rechenberg, I., 1989.
    Evolution strategy: Nature’s way of optimization.
    In Optimization: Methods and Applications, Possibilities and Limitations (pp. 106-126).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-83814-9_6

    Schwefel, H.P., 1984.
    Evolution strategies: A family of non-linear optimization techniques based on imitating
    some principles of organic evolution.
    Annals of Operations Research, 1(2), pp.165-167.
    https://link.springer.com/article/10.1007/BF01876146

    Rechenberg, I., 1984.
    The evolution strategy. A mathematical model of darwinian evolution.
    In Synergetics—from Microscopic to Macroscopic Order (pp. 122-132). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-642-69540-7_13
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of offspring (λ: lambda), offspring population size
            self.n_individuals = 4 + int(3*np.log(self.ndim_problem))  # for small populations setting
        assert self.n_individuals > 0, f'`self.n_individuals` = {self.n_individuals}, but should > 0.'
        if self.n_parents is None:  # number of parents (μ: mu), parental population size
            self.n_parents = int(self.n_individuals / 2)
            if self.n_parents > 1:
                # for consistency and simplicity, we unify these in the base class ES.
                w_base, w = np.log((self.n_individuals + 1) / 2), np.log(np.arange(self.n_parents) + 1)
                self._w = (w_base - w) / (self.n_parents*w_base - np.sum(w))
                self._mu_eff = 1 / np.sum(np.power(self._w, 2))  # μ_eff / μ_w
                # E[||N(0,I)||]: expectation of chi distribution
                self._e_chi = np.sqrt(self.ndim_problem)*(
                        1 - 1 / (4*self.ndim_problem) + 1 / (21*np.power(self.ndim_problem, 2)))
        assert self.n_parents > 0, f'`` = {self.n_parents}, but should > 0.'
        self.mean = options.get('mean')  # mean of Gaussian search distribution
        if self.mean is None:  # 'mean' has priority over 'x'
            self.mean = options.get('x')
        self.sigma = options.get('sigma')  # global step-size (σ)
        self.eta_mean = options.get('eta_mean')  # learning rate of mean
        self.eta_sigma = options.get('eta_sigma')  # learning rate of std
        self._n_generations = 0
        # for restart
        self.n_restart = 0
        self.sigma_threshold = options.get('sigma_threshold', 1e-10)  # stopping threshold of sigma for restart
        # maximal generation number of fitness stagnation for restart
        self.stagnation = options.get('stagnation', np.maximum(32, self.ndim_problem))
        self.fitness_diff = options.get('fitness_diff', 1e-10)  # stopping threshold of fitness difference for restart
        self._sigma_bak = np.copy(self.sigma)  # bak for restart
        self._fitness_list = [self.best_so_far_y]  # to store `best_so_far_y` generated in each generation

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _initialize_mean(self, is_restart=False):
        if is_restart or (self.mean is None):
            mean = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            mean = np.copy(self.mean)
        return mean

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose_frequency):
            best_so_far_y = -self.best_so_far_y if self._is_maximization else self.best_so_far_y
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, best_so_far_y, np.min(y), self.n_function_evaluations))

    def restart_initialize(self):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            self.n_individuals *= 2
            self.n_parents = int(self.n_individuals / 2)
            if self.n_parents > 1:
                w_base, w = np.log((self.n_individuals + 1) / 2), np.log(np.arange(self.n_parents) + 1)
                self._w = (w_base - w) / (self.n_parents*w_base - np.sum(w))
                self._mu_eff = 1 / np.sum(np.power(self._w, 2))
            self._n_generations = 0
            self._fitness_list = [np.Inf]
        return is_restart

    def _collect_results(self, fitness, mean=None):
        results = Optimizer._collect_results(self, fitness)
        results['mean'] = mean
        results['sigma'] = self.sigma
        results['n_restart'] = self.n_restart
        results['_n_generations'] = self._n_generations
        return results



import time
import numpy as np

from pypop7.benchmarks import rotated_functions as rf
from pypop7.benchmarks.rotated_functions import generate_rotation_matrix


def generate_rotation_matrices(functions=None, ndims=None, seed=None):
    if functions is None:
        functions = [rf.sphere, rf.cigar, rf.discus, rf.cigar_discus, rf.ellipsoid,
                     rf.different_powers, rf.schwefel221, rf.step, rf.rosenbrock, rf.schwefel12]
    if ndims is None:
        ndims = [2, 10, 100, 200, 1000, 2000]
    if seed is None:
        seed = 2022

    rng = np.random.default_rng(seed)
    seeds = rng.integers(np.iinfo(np.int64).max, size=(len(functions), len(ndims)))

    start_run = time.time()
    for i, f in enumerate(functions):
        for j, d in enumerate(ndims):
            start_time = time.time()
            generate_rotation_matrix(f, d, seeds[i, j])
            print('* {:d}-d {:s}: runtime {:7.5e}'.format(
                d, f.__name__, time.time() - start_time))
    print('*** Total runtime: {:7.5e}.'.format(time.time() - start_run))


if __name__ == '__main__':
    generate_rotation_matrices()



from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.res import RES
from pypop7.optimizers.es.ssaes import SSAES
from pypop7.optimizers.es.dsaes import DSAES
from pypop7.optimizers.es.csaes import CSAES
from pypop7.optimizers.es.saes import SAES


__all__ = [ES, RES, SSAES, DSAES, CSAES, SAES]



import numpy as np

from pypop7.optimizers.es.lmcmaes import LMCMAES


class LMCMA(LMCMAES):
    """Limited Memory Covariance Matrix Adaptation (LMCMA, (μ/μ_w,λ)-LM-CMA).

    Reference
    ---------
    Loshchilov, I., 2017.
    LM-CMA: An alternative to L-BFGS for large-scale black box optimization.
    Evolutionary Computation, 25(1), pp.143-171.
    https://direct.mit.edu/evco/article-abstract/25/1/143/1041/LM-CMA-An-Alternative-to-L-BFGS-for-Large-Scale
    (See Algorithm 7 for details.)

    See the official C++ version from Loshchilov:
    https://sites.google.com/site/ecjlmcma/
    """
    def __init__(self, problem, options):
        LMCMAES.__init__(self, problem, options)
        self.base_m = options.get('base_m', 4)  # base number of direction vectors
        self.period = options.get('period', np.maximum(1, int(np.log(self.ndim_problem))))  # update period
        # self.z_star = options.get('z_star', 0.25)  # target success rate for PSR

    def _a_z(self, z=None, pm=None, vm=None, b=None, start=None, it=None):
        x = np.copy(z)
        for t in range(start, it):
            x = self._a * x + b[self._j[t]] * np.dot(vm[self._j[t]], z) * pm[self._j[t]]
        return x

    def _rademacher(self):
        random = self.rng_optimization.integers(2, size=(self.ndim_problem,))
        random[random == 0] = -1
        return random

    def iterate(self, mean=None, x=None, pm=None, vm=None, y=None, b=None, args=None):
        it = int(np.minimum(self.m, (self._n_generations / self.period)))
        sign, z = 1, np.empty((self.ndim_problem,))  # for mirrored sampling
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            if sign == 1:
                base_m = (10 * self.base_m if k == 0 else self.base_m) * np.abs(
                    self.rng_optimization.standard_normal())
                base_m = it if base_m > it else base_m
                z = self._a_z(self._rademacher(), pm, vm, b, int(it - base_m if it > 1 else 0), it)
            x[k] = mean + sign * self.sigma * z
            y[k] = self._evaluate_fitness(x[k], args)
            sign *= -1  # sample in the opposite direction for mirrored sampling
        return x, y

    def _update_distribution(self, mean=None, x=None, p_c=None, s=None, vm=None, pm=None,
                             b=None, d=None, y=None, y_bak=None):
        mean_bak = np.dot(self._w, x[np.argsort(y)[:self.n_parents]])
        p_c = self._p_c_1 * p_c + self._p_c_2 * (mean_bak - mean) / self.sigma
        if self._n_generations % self.period == 0:
            _n_generations = int(self._n_generations / self.period)
            if _n_generations < self.m:
                i_min, self._j[_n_generations] = _n_generations, _n_generations
            elif self.m > 1:
                i_min, d_min = 1, self._l[self._j[1]] - self._l[self._j[0]]
                for j in range(2, self.m):
                    d_cur = self._l[self._j[j]] - self._l[self._j[j - 1]]
                    if d_cur < d_min:
                        d_min, i_min = d_cur, j
                # if all pairwise distances exceed self.n_steps, start from 0
                i_min = 0 if d_min >= self.n_steps else i_min
                updated = self._j[i_min]
                for j in range(i_min, self.m - 1):
                    self._j[j] = self._j[j + 1]
                self._j[self.m - 1] = updated
            else:
                i_min = 0
            pm[self._j[np.minimum(self.m - 1, _n_generations)]] = p_c
            self._l[self._j[np.minimum(self.m - 1, _n_generations)]] = self._n_generations
            for i in range(i_min, np.minimum(self.m, _n_generations + 1)):
                vm[self._j[i]] = self._a_inv_z(pm[self._j[i]], vm, d, i)
                v_n = np.dot(vm[self._j[i]], vm[self._j[i]])
                bd_3 = np.sqrt(1 + self._bd_2 * v_n)
                b[self._j[i]] = self._bd_1 / v_n * (bd_3 - 1)
                d[self._j[i]] = 1 / (self._bd_1 * v_n) * (1 - 1 / bd_3)
        y.sort()
        if self._n_generations > 0:
            r = np.argsort(np.hstack((y, y_bak)))
            z_psr = np.sum(self._rr[r < self.n_individuals] - self._rr[r >= self.n_individuals])
            z_psr = z_psr / np.power(self.n_individuals, 2) - self.z_star
            s = (1 - self.c_s) * s + self.c_s * z_psr
            self.sigma *= np.exp(s / self.d_s)
        return mean_bak, p_c, s, vm, pm, b, d



"""Repeat Fig. 4 (Ellipsoid N=3 and N=30) from the following paper:
    Beyer, H.G. and Sendhoff, B., 2017.
    Simplify your covariance matrix adaptation evolution strategy.
    IEEE Transactions on Evolutionary Computation, 21(5), pp.746-759.
    https://ieeexplore.ieee.org/document/7875115
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from optimizers.es.es import ES
from optimizers.es.maes import MAES
from pypop7.benchmarks.base_functions import ellipsoid

sns.set_theme(style='darkgrid')


class Fig4(MAES):
    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fit = ES.optimize(self, fitness_function)
        z, d, mean, s, tm, y = self.initialize()
        f, stepsize = [], []  # for plotting data
        while True:
            z, d, y = self.iterate(z, d, mean, tm, y, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fit.extend(y)
            f.append(np.min(y))
            stepsize.append(self.sigma)
            if self._check_terminations():
                break
            mean, s, tm = self._update_distribution(z, d, mean, s, tm, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            z, d, mean, s, tm, y = self.restart_initialize(z, d, mean, s, tm, y)
        res = self._collect_results(fit, mean)
        res['s'] = s
        res['f'] = f
        res['stepsize'] = stepsize
        return res


if __name__ == '__main__':
    for ndim_problem in [3, 30]:
        problem = {'fitness_function': ellipsoid,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -10 * np.ones((ndim_problem,)),
                   'upper_boundary': 10 * np.ones((ndim_problem,))}
        options = {'fitness_threshold': 1e-15,
                   'seed_rng': 2022,  # not given in the original paper
                   'x': np.ones((ndim_problem,)),
                   'sigma': 1.0,
                   'stagnation': np.Inf}
        maes = Fig4(problem, options)
        results = maes.optimize()
        plt.figure()
        plt.yscale('log')
        plt.yticks(ticks=[1e-15, 1e-10, 1e-5, 1e0, 1e5])
        plt.plot(np.arange(len(results['f'])) + 1, results['f'], 'r-')
        plt.plot(np.arange(len(results['stepsize'])) + 1, results['stepsize'], 'm-')
        plt.xlabel('g')
        plt.ylabel('sigma and f(y)')
        plt.title('Ellipsoid N = ' + str(ndim_problem))
        plt.show()



import numpy as np

from optimizers.pso.pso import PSO


class OPSO(PSO):
    """Online Particle Swarm Optimizer (OPSO).

    Reference
    ---------
    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (Vol. 4, pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146

    MATLAB Source Code:
    https://github.com/P-N-Suganthan/CODES/blob/master/2006-IEEE-TEC-CLPSO.zip
    """
    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)

    def initialize(self, args=None):
        x, y, p_x, p_y, n_x, v = PSO.initialize(self)
        self.record_fitness_initialization = True  # evaluate fitness in advance
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y, p_x, p_y, n_x, v
            y[i] = self._evaluate_fitness(x[i], args)
        p_y = np.copy(y)
        self.n_generations += 1
        return x, y, p_x, p_y, n_x, v

    def iterate(self, x=None, y=None, p_x=None, p_y=None, n_x=None, v=None, args=None):  # use online update
        rng = self.rng_optimization
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y, p_x, p_y, n_x, v
            # evaluate fitness
            y[i] = self._evaluate_fitness(x[i], args)
            if y[i] < p_y[i]:
                p_x[i], p_y[i] = x[i], y[i]
            # update neighbor topology
            n_x[i], _ = self.topology(p_x, p_y, i)
            # update and limit velocity
            cognition_rand = rng.uniform(size=(self.ndim_problem,))
            society_rand = rng.uniform(size=(self.ndim_problem,))
            v[i] = self.w[self.n_generations] * v[i] +\
                self.cognition * cognition_rand * (p_x[i] - x[i]) +\
                self.society * society_rand * (n_x[i] - x[i])
            less_min_v, more_max_v = v[i] < self.min_v, v[i] > self.max_v
            v[i, less_min_v], v[i, more_max_v] = self.min_v[less_min_v], self.max_v[more_max_v]
            # update position
            x[i] += v[i]
        return x, y, p_x, p_y, n_x, v



import numpy as np

from pypop7.optimizers.es.es import ES


class LMCMAES(ES):
    """Limited Memory Covariance Matrix Adaptation Evolution Strategy (LMCMAES, (μ/μ_w,λ)-LM-CMA-ES).

    Reference
    ---------
    Loshchilov, I., 2014, July.
    A computationally efficient limited memory CMA-ES for large scale optimization.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 397-404). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598294
    (See Algorithm 6 for details.)

    See the official C++ version from Loshchilov:
    https://sites.google.com/site/lmcmaeses/
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_s = options.get('c_s', 0.3)  # learning rate for population success rule (PSR)
        self.d_s = options.get('d_s', 1)
        self.m = options.get('m', self.n_individuals)  # number of direction vectors
        self.n_steps = options.get('n_steps', self.m)  # target number of generations between vectors
        self.c_c = options.get('c_c', 1 / self.m)  # learning rate for evolution path
        # learning rate for covariance matrix adaptation
        self.c_1 = options.get('c_1', 1 / (10 * np.log(self.ndim_problem + 1)))
        self.z_star = options.get('z_star', 0.25)  # target success rate for PSR
        self._a = np.sqrt(1 - self.c_1)  # for Algorithm 3
        self._c = 1 / np.sqrt(1 - self.c_1)  # for Algorithm 4
        self._bd_1 = np.sqrt(1 - self.c_1)  # for Line 13 and 14
        self._bd_2 = self.c_1 / (1 - self.c_1)  # for Line 13 and 14
        self._p_c_1 = 1 - self.c_c  # for Line 9
        self._p_c_2 = None  # for Line 9
        self._rr = None  # for PSR
        self._j = None
        self._l = None

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        s = 0  # for PSR of global step-size adaptation
        vm = np.empty((self.m, self.ndim_problem))
        pm = np.empty((self.m, self.ndim_problem))
        b = np.empty((self.m,))
        d = np.empty((self.m,))
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._p_c_2 = np.sqrt(self.c_c * (2 - self.c_c) * self._mu_eff)
        self._rr = np.arange(self.n_individuals * 2, 0, -1) - 1
        self._j = [None] * self.m
        self._l = [None] * self.m
        return mean, x, p_c, s, vm, pm, b, d, y

    def _a_z(self, z=None, pm=None, vm=None, b=None):  # Cholesky factor - vector update
        # Algorithm 3 Az()
        x = np.copy(z)
        for t in range(np.minimum(self.m, self._n_generations)):
            x = self._a * x + b[self._j[t]] * np.dot(vm[self._j[t]], z) * pm[self._j[t]]
        return x

    def iterate(self, mean=None, x=None, pm=None, vm=None, y=None, b=None, args=None):
        for k in range(self.n_individuals):  # Line 4
            if self._check_terminations():
                return x, y
            z = self.rng_optimization.standard_normal((self.ndim_problem,))  # Line 5
            x[k] = mean + self.sigma * self._a_z(z, pm, vm, b)  # Line 6
            y[k] = self._evaluate_fitness(x[k], args)  # Line 7
        return x, y

    def _a_inv_z(self, v=None, vm=None, d=None, i=None):  # inverse Cholesky factor - vector update
        # Algorithm 4 Ainvz()
        x = np.copy(v)
        for t in range(0, i):
            x = self._c * x - d[self._j[t]] * np.dot(vm[self._j[t]], x) * vm[self._j[t]]
        return x

    def _update_distribution(self, mean=None, x=None, p_c=None, s=None, vm=None, pm=None,
                             b=None, d=None, y=None, y_bak=None):
        mean_bak = np.dot(self._w, x[np.argsort(y)[:self.n_parents]])  # Line 8
        p_c = self._p_c_1 * p_c + self._p_c_2 * (mean_bak - mean) / self.sigma  # Line 9
        if self._n_generations < self.m:  # before self.m generations
            self._j[self._n_generations] = self._n_generations
            i_min = self._n_generations
        else:
            i_min = 1  # start from the latter (1) of pairwise evolution paths
            d_min = self._l[self._j[i_min]] - self._l[self._j[i_min - 1]]
            for j in range(2, self.m):
                d_cur = self._l[self._j[j]] - self._l[self._j[j - 1]]
                if d_cur < d_min:
                    d_min, i_min = d_cur, j
            # if all pairwise distances exceed self.n_steps, start from 0
            i_min = 0 if d_min >= self.n_steps else i_min
            # update indexes of evolution paths (self._j[i_min] is index of evolution path needed to delete)
            updated = self._j[i_min]
            for j in range(i_min, self.m - 1):
                self._j[j] = self._j[j + 1]
            self._j[self.m - 1] = updated
        # note that the latest evolution path is stored in self._j[np.minimum(self.m - 1, self._n_generations)]
        pm[self._j[np.minimum(self.m - 1, self._n_generations)]] = p_c  # add the latest evolution path
        self._l[self._j[np.minimum(self.m - 1, self._n_generations)]] = self._n_generations  # update its generation
        # since self._j[i_min] is deleted, all vectors (from vm) depending on it need to be computed again
        for i in range(i_min, np.minimum(self.m, self._n_generations + 1)):
            vm[self._j[i]] = self._a_inv_z(pm[self._j[i]], vm, d, i)
            v_n = np.dot(vm[self._j[i]], vm[self._j[i]])
            bd_3 = np.sqrt(1 + self._bd_2 * v_n)
            b[self._j[i]] = self._bd_1 / v_n * (bd_3 - 1)
            d[self._j[i]] = 1 / (self._bd_1 * v_n) * (1 - 1 / bd_3)
        y.sort()
        if self._n_generations > 0:
            r = np.argsort(np.hstack((y, y_bak)))  # for Line 15
            z_psr = np.sum(self._rr[r < self.n_individuals] - self._rr[r >= self.n_individuals])  # Line 15
            z_psr = z_psr / np.power(self.n_individuals, 2) - self.z_star  # Line 15
            s = (1 - self.c_s) * s + self.c_s * z_psr  # Line 17
            self.sigma *= np.exp(s / self.d_s)  # Line 18
        return mean_bak, p_c, s, vm, pm, b, d

    def restart_initialize(self, args=None, mean=None, x=None, p_c=None, s=None, vm=None, pm=None,
                           b=None, d=None, y=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            mean, x, p_c, s, vm, pm, b, d, y = self.initialize(is_restart)
            self.d_s *= 2
        return mean, x, p_c, s, vm, pm, b, d, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, x, p_c, s, vm, pm, b, d, y = self.initialize(args)
        while True:
            y_bak = np.copy(y)
            x, y = self.iterate(mean, x, pm, vm, y, b, args)  # sample and evaluate offspring population
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p_c, s, vm, pm, b, d = self._update_distribution(mean, x, p_c, s, vm, pm, b, d, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, x, p_c, s, vm, pm, b, d, y = self.restart_initialize(args, mean, x, p_c, s, vm, pm, b, d, y)
        results = self._collect_results(fitness, mean)
        results['p_c'] = p_c
        results['s'] = s
        return results



"""Repeat Fig.1 in this paper
    (1+1)-Active-CMA-ES (OPOA).
    Reference
    ---------
    Arnold, D.V. and Hansen, N., 2010, July.
    Active covariance matrix adaptation for the (1+1)-CMA-ES.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 385-392). ACM.
    https://dl.acm.org/doi/abs/10.1145/1830483.1830556
    """
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle
import time

from pypop7.optimizers.es.opoa import OPOA
from pypop7.optimizers.es.es import ES
from pypop7.benchmarks.base_functions import discus

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function, problem_dim):
    result = read_pickle(function, problem_dim)
    index = np.arange(1, len(result['sigmas']) + 1, 1)
    plt.figure()
    plt.yscale('log')
    plt.ylim([1e-9, 1e6])
    plt.yticks([1e-9, 1e-6, 1e-3, 1e0, 1e3, 1e6])
    plt.xticks([0, 1e3, 2e3, 3e3, 4e3, 5e3, 6e3])
    plt.xlim([0, 6e3])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='r', label='function value f(x)')
    plt.plot(index, result['sigmas'], color='g', label=r'mutation strength $\sigma$')
    plt.xlabel("function evaluations")
    plt.ylabel("function value")
    plt.title(function.capitalize() + " Dimension: " + str(problem_dim))
    plt.legend()
    plt.show()


class Fig1(OPOA):
    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        sigmas = []
        start_run = time.time()
        mean, y, a, a_i, best_so_far_y, p_s, p_c = self.initialize(args)
        fitness.append(y)
        while True:
            mean, y, a, a_i, best_so_far_y, p_s, p_c = self.iterate(
                args, mean, a, a_i, best_so_far_y, p_s, p_c)
            sigmas.append(self.sigma)
            if self.record_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, y, a, a_i, best_so_far_y, p_s, p_c = self.restart_initialize(
                    args, mean, y, a, a_i, best_so_far_y, p_s, p_c, fitness)
        results = self._collect_results(fitness, mean)
        results['runtime'] = time.time() - start_run
        results['sigmas'] = sigmas
        return results


if __name__ == '__main__':
    for f in [discus]:
        for d in [10]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 5 * np.ones((d,))}
            options = {'max_function_evaluations': 2e6,
                       'fitness_threshold': 1e-9,
                       'max_runtime': 3600,  # 1 hours
                       'seed_rng': 0,
                       'x': 4 * np.ones((d,)),  # mean
                       'sigma': 0.1,
                       'verbose_frequency': 2000,
                       'record_fitness': True,
                       'is_restart': False,
                       'record_fitness_frequency': 1}
            solver = Fig1(problem, options)
            results = solver.optimize()
            write_pickle(f.__name__, d, results)
            plot(f.__name__, d)
